# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
import os
import logging

from future.builtins import str, int
import pytest
import numpy as N

from astrophysix.simdm import Project, ProjectCategory, SimulationStudy
from astrophysix.simdm.experiment import Simulation
from astrophysix.simdm.protocol import SimulationCode
from astrophysix.simdm.results import GenericResult, Snapshot
from astrophysix.simdm.catalogs import Catalog, TargetObject, ObjectProperty, CatalogField, ObjectPropertyGroup
from astrophysix.simdm.datafiles import Datafile
from astrophysix.simdm.services import CatalogDataProcessingService, CatalogFieldBinding
from astrophysix import units as U
from astrophysix.simdm.utils import DataType

from astrophysix.utils import Hdf5StudyPersistent, Hdf5StudyFileMode


class TestCatalogField(object):
    def test_catalog_field_init(self):
        """
        Tests object catalog field instance initialisation
        """
        # Catalog field initialisation
        tobj = TargetObject(name="Elliptic galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        cat = Catalog(target_object=tobj, name="Elliptic galaxy catalog")
        fx = cat.catalog_fields.add(CatalogField(x, values=N.random.uniform(size=10)))
        fy = cat.catalog_fields.add(CatalogField(y, values=N.random.uniform(size=10)))

        assert fx.property_name == x.property_name
        assert fy.object_property is y
        assert fx.nobjects == cat.nobjects == 10

        # Test str() conversion
        assert str(fx) == "'pos_x' catalog field"

    def test_catalog_field_values_exc(self):
        """
        Tests that a catalog field set with empty value array or array with dimensions != 1 raises an error
        """
        y = ObjectProperty(property_name="pos_y")

        # Undefined value array => error
        with pytest.raises(AttributeError, match="Undefined 'values' property in CatalogField."):
            f = CatalogField(y)

        # Empty array => error
        with pytest.raises(AttributeError, match="CatalogField 'values' array need at least 1 value."):
            f = CatalogField(y, values=N.array([]))

        # 2D array => error
        with pytest.raises(AttributeError, match="'values' CatalogField attribute must be a 1-dimensional array."):
            f = CatalogField(y, values=N.random.uniform(size=(10, 5)))

        # Ok => 10 items in fields
        fy = CatalogField(y, values=N.random.uniform(size=10))

        # Tries to change the size of a CatalogField => error
        with pytest.raises(AttributeError, match="The number of items in this CatalogField is 10 and cannot be changed."):
            fy.field_values = N.random.uniform(size=12)

    def test_catalog_field_object_property_setting_exc(self):
        """
        Tests setting an ObjectProperty into a CatalogField object
        """
        tobj = TargetObject(name="My object")
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))

        # Tests CatalogField defined without target object property
        with pytest.raises(AttributeError, match="Undefined 'obj_prop' object property attribute in CatalogField."):
            CatalogField()  # No target object property defined

        # Tests invalid target object property initialisation exception
        invalid_objprop_err = "CatalogField 'obj_prop' attribute is not a valid ObjectProperty instance."
        with pytest.raises(AttributeError, match=invalid_objprop_err):
            CatalogField(obj_prop=-5)  # Invalid target object property
        with pytest.raises(AttributeError, match=invalid_objprop_err):
            CatalogField("Test")  # Invalid target object property

        # Catalog field initialisation : valid => should not raise any exception
        f1 = CatalogField(obj_prop=y, values=N.random.uniform(size=5))
        f2 = CatalogField(y, values=N.random.uniform(size=5))

    def test_catalog_field_equality(self):
        """
        Tests catalog field rich comparison method CatalogField.__eq__()
        """
        x = ObjectProperty(property_name="pos_x")
        y = ObjectProperty(property_name="pos_y")
        z = ObjectProperty(property_name="pos_z")
        cfx = CatalogField(x, values=N.random.uniform(size=20))

        # Different UUID => not equals
        assert cfx != CatalogField(cfx.object_property, values=cfx.field_values)

        # Different object property => not equals
        assert cfx != CatalogField(y, values=cfx.field_values, uid=cfx.uid)

        # Different values => not equals
        assert cfx != CatalogField(cfx.object_property, values=N.random.uniform(size=20), uid=cfx.uid)

        # Identical catalog fields
        f = CatalogField(cfx.object_property, values=cfx.field_values, uid=cfx.uid)
        assert cfx == f

    def test_catalog_field_to_pandas(self):
        """Tests CatalogField conversion into a pandas.Series"""
        x = ObjectProperty(property_name="pos_x", unit=U.kpc)
        cf = CatalogField(obj_prop=x, values=N.random.uniform(size=200))
        s = cf.to_pandas()
        assert s.size == 200
        assert s.name == "pos_x (\\textrm{kpc})"

    def test_catalog_field_hdf5_io(self, tmp_path):
        """
        Tests saving/loading CatalogField from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        # Add tabulated values in catalog + properties in target object
        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        z = tobj.object_properties.add(ObjectProperty(property_name="pos_z"))
        cat = Catalog(target_object=tobj, name="Spiral galaxy catalog", description="My descr.")
        vals_x = N.random.uniform(size=200)
        cfx = cat.catalog_fields.add(CatalogField(obj_prop=x, values=vals_x))

        res = GenericResult(name="Key result")
        res.catalogs.add(cat)

        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.generic_results.add(res)

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]

        # Compare catalog fields
        res_loaded = simu_loaded.generic_results[res.name]
        cat_loaded = res_loaded.catalogs[cat.name]
        field_loaded = cat_loaded.catalog_fields[cfx.property_name]
        assert cfx == field_loaded and field_loaded.nobjects == 200

        cfx_min, cfx_max, cfx_mean, cfx_std = field_loaded.field_value_stats
        assert cfx_min == vals_x.min() and cfx_max == vals_x.max()
        assert cfx_mean == vals_x.mean() and cfx_std == vals_x.std()


class TestCatalog(object):
    def test_catalog_init(self):
        """
        Tests object catalog instance initialisation
        """
        # Object catalog initialisation
        tobj = TargetObject(name="Elliptic galaxy")
        cat = Catalog(target_object=tobj, name="Elliptic galaxy catalog",
                      description="This is the catalog of all elliptic galaxies identified in the simulation")

        assert cat.name == "Elliptic galaxy catalog"
        assert cat.target_object is tobj
        assert cat.description == "This is the catalog of all elliptic galaxies identified in the simulation"

        # Test str() conversion
        assert str(cat) == "'Elliptic galaxy catalog' catalog"

    def test_catalog_targetobject_setting_exc(self):
        """
        Tests setting a TargetObject into a Catalog object
        """
        # Tests Catalog defined without target object
        with pytest.raises(AttributeError, match="Undefined target object for 'My super catalog' catalog."):
            cat_exc = Catalog(name="My super catalog")  # No target object defined

        # Tests invalid target_object initialisation exception
        invalid_targobj_err = "Catalog 'target_object' attribute is not a valid TargetObject instance."
        with pytest.raises(AttributeError, match=invalid_targobj_err):
            cat_exc = Catalog(name="My catalog", target_object=-5)  # Invalid target object
        with pytest.raises(AttributeError, match=invalid_targobj_err):
            cat_exc = Catalog(-5, name="My catalog")  # Invalid target object

        # Catalog initialisation : valid => should not raise any exception
        core = TargetObject(name="Pre-stellar core")
        cat1 = Catalog(core, name="My pre-stellar core catalog")
        cat2 = Catalog(name="My pre-stellar core catalog", target_object=core)

    def test_setting_catalog_name_exc(self):
        """
        Tests that a catalog defined without name raises an exception
        """
        # Tests catalog defined without name
        with pytest.raises(AttributeError) as e:
            cat_exc = Catalog()  # No catalog name defined
        assert str(e.value) == "Catalog 'name' attribute is not defined (mandatory)."

        cat = Catalog(target_object=TargetObject(name="Pre-stellar core"), name="My catalog")

        # ------------------- Tests invalid catalog name setting exception ----------------------- #
        empty_name_err = "Catalog 'name' property is not a valid (non-empty) string."
        with pytest.raises(AttributeError) as e:
            cat.name = -1
        assert str(e.value) == empty_name_err

        with pytest.raises(AttributeError) as e:
            cat.name = ""
        assert str(e.value) == empty_name_err
        # ---------------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        cat.name = "My best catalog"

    def test_setting_catalog_description(self):
        """
        Tests setting catalog description property
        """
        cat = Catalog(target_object=TargetObject(name="Pre-stellar core"), name="My super catalog")

        # ------------------------------ Tests invalid catalog description setting exception ------------------------- #
        with pytest.raises(AttributeError) as e:
            cat.description = 0
        assert str(e.value) == "Catalog 'description' property is not a valid string."

        # Valid => should not raise any exception
        cat.description = "My catalog description"
        # ------------------------------------------------------------------------------------------------------------ #

    def test_catalog_field_binding_insertion(self):
        """
        Tests exceptions raised upon object catalog field binding insertion into Catalog instances
        """
        # Create TargetObject + properties
        tobj = TargetObject(name="Spiral galaxy", description="This is a disk-like galaxy")
        alpha = tobj.object_properties.add(ObjectProperty(property_name='alpha'))
        beta = tobj.object_properties.add(ObjectProperty(property_name='beta'))
        gamma = tobj.object_properties.add(ObjectProperty(property_name='gamma'))

        # Create catalog + fields
        cat = Catalog(target_object=tobj, name="Elliptic galaxy catalog")
        alpha_f = CatalogField(alpha, values=N.random.uniform(size=100))
        beta_f = cat.catalog_fields.add(CatalogField(beta, values=N.random.uniform(size=100)))
        gamma_f = cat.catalog_fields.add(CatalogField(gamma, values=N.random.uniform(size=100)))

        # Create data processing service
        serv = cat.processing_services.add(CatalogDataProcessingService(service_name="my_service", data_host="IClust"))
        fb_alpha = CatalogFieldBinding(catalog_field=alpha_f, param_key="alpha")

        # Tries to add catalog field binding to the service before the field has been added to the catalog field
        # list => raises an error
        with pytest.raises(AttributeError, match="Cannot add catalog field binding in this data processing service, "
                                                 "{f!s} is not a field of the {cat!s}.".format(f=alpha_f, cat=cat)):
            serv.catalog_field_bindings.add(fb_alpha)

        # Tries to add a data processing service into a catalog while one of its field binding refers to a catalog field
        # that does not belong to the catalog field list (yet) => raises an error
        serv2 = CatalogDataProcessingService(service_name="my_service2", data_host="IClust")
        serv2.catalog_field_bindings.add(fb_alpha)
        fb_beta = serv2.catalog_field_bindings.add(CatalogFieldBinding(catalog_field=beta_f, param_key="beta"))
        fb_gamma = serv2.catalog_field_bindings.add(CatalogFieldBinding(catalog_field=gamma_f, param_key="gamma"))
        with pytest.raises(AttributeError, match="Cannot add catalog processing service. {f!s} is not a field of "
                                                 "the {cat!s}.".format(f=alpha_f, cat=cat)):
            cat.processing_services.add(serv2)

        # # Valid => no error
        cat.catalog_fields.add(alpha_f)
        cat.processing_services.add(serv2)  # Ok
        serv.catalog_field_bindings.add(fb_alpha)  # Ok

    def test_deleting_catfield_from_catalog_while_referenced_in_binding(self):
        """
        Tests object property deletion from TargetObject's property list while object property is included in one of the
        TargetObject's ObjectPropertyGroup.
        """
        # Create TargetObject + properties
        tobj = TargetObject(name="Spiral galaxy", description="This is a disk-like galaxy")
        alpha = tobj.object_properties.add(ObjectProperty(property_name='alpha'))

        # Create catalog + fields
        cat = Catalog(target_object=tobj, name="Elliptic galaxy catalog")
        alpha_f = cat.catalog_fields.add(CatalogField(alpha, values=N.random.uniform(size=100)))

        # Create data processing service
        serv = cat.processing_services.add(CatalogDataProcessingService(service_name="Best service", data_host="IClust"))
        fb_alpha = serv.catalog_field_bindings.add(CatalogFieldBinding(catalog_field=alpha_f, param_key="alpha"))

        # Tries to delete 'alpha' catalog field, while it is referenced in a binding of the 'Best service' service
        with pytest.raises(AttributeError, match="'{f!s}' cannot be deleted, the following items depend on it \\(try to"
                                                 " delete them first\\) : \\[{fb!s}\\].".format(f=alpha_f,
                                                                                                fb=fb_alpha)):
            del cat.catalog_fields[alpha_f.property_name]

        # Delete 'alpha' field binding from 'Best service' data processing service first
        del serv.catalog_field_bindings[fb_alpha.field_property_name]

        # THEN delete the 'alpha' field from the catalog => Ok
        del cat.catalog_fields[alpha_f.property_name]

    def test_catalog_equality(self):
        """
        Tests catalog rich comparison method Catalog.__eq__()
        """
        tobj = TargetObject(name="Elliptic galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        z = tobj.object_properties.add(ObjectProperty(property_name="pos_z"))
        cat = Catalog(tobj, name="Elliptic galaxy catalog", description="This is my elliptic galaxy catalog")

        # Different UUID => not equals
        assert cat != Catalog(tobj, name=cat.name, description=cat.description)

        # Different target object => not equals
        assert cat != Catalog(TargetObject(name="Spiral galaxy"), name=cat.name, description=cat.description,
                              uid=cat.uid)

        # Different name => not equals
        assert cat != Catalog(tobj, name="Name", description=cat.description, uid=cat.uid)

        # Different description => not equals
        assert cat != Catalog(tobj, name=cat.name, description="Alt. description", uid=cat.uid)

        # Identical catalogs
        c = Catalog(tobj, name=cat.name, description=cat.description, uid=cat.uid)
        assert cat == c

        # Add datafiles
        df1 = cat.datafiles.add(Datafile(name="My important datafile"))
        df2 = cat.datafiles.add(Datafile(name="My important datafile (2)"))
        # Datafile list differ => not equals
        c.datafiles.add(df1)
        assert c != cat
        # Identical catalogs
        c.datafiles.add(df2)
        assert c == cat

        # Compare catalog tabulated values + fields
        cfx = cat.catalog_fields.add(CatalogField(x, values=N.random.uniform(size=10)))
        cfy = cat.catalog_fields.add(CatalogField(y, values=N.random.uniform(size=10)))
        # Catalog field list differ => not equals
        c.catalog_fields.add(cfx)
        assert cat != c
        # Identical catalogs
        c.catalog_fields.add(cfy)
        assert cat == c

        # Compare catalog data processing services
        cdps = cat.processing_services.add(CatalogDataProcessingService(service_name="surface_density_map",
                                                                        data_host="IClust"))
        # Catalog data processing service list differ => not equals
        assert cat != c
        # Identical catalogs
        c.processing_services.add(cdps)
        assert cat == c

    def test_catalog_with_incompatible_fields(self):
        """
        Tests that adding two fields with different number of items in a catalog raises an Exception
        """
        tobj = TargetObject(name="Pre-stellar core")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        cat = Catalog(target_object=tobj, name="Elliptic galaxy catalog")
        cat.catalog_fields.add(CatalogField(x, values=N.random.uniform(size=10)))
        with pytest.raises(AttributeError, match="Cannot add a catalog field with 12 item values in a catalog "
                                                 "containing 10 items."):
            cat.catalog_fields.add(CatalogField(y, values=N.random.uniform(size=12)))

    def test_catalog_adding_field_with_unknown_targobj_property(self):
        """
        Tests that adding a field linked to a property that does not belong to the catalog's target object property list
        raises an Exception
        """
        tobj = TargetObject(name="Pre-stellar core")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        z = ObjectProperty(property_name="pos_z")  # z not added in target object
        cat = Catalog(target_object=tobj, name="Elliptic galaxy catalog")

        # Tries to add a catalog field with z property => error
        with pytest.raises(AttributeError, match="Cannot add catalog field. 'pos_z' target object property is not a "
                                                 "property of 'Pre-stellar core' target object."):
            cat.catalog_fields.add(CatalogField(z, values=N.random.uniform(size=5)))

    def test_deleting_object_prop_from_catalog_targobj_while_linked_to_field(self):
        """
        Tests object property deletion from catalog's target object while object property is linked to a catalog field.
        """
        tobj = TargetObject(name="Pre-stellar core")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        cat = Catalog(target_object=tobj, name="Elliptic galaxy catalog")
        fx = cat.catalog_fields.add(CatalogField(x, values=N.random.uniform(size=10)))

        # Tries to delete x, while it is linked to fx catalog field
        with pytest.raises(AttributeError, match="''pos_x' target object property' cannot be deleted, the following "
                                                 "items depend on it \\(try to delete them first\\) : \\['Elliptic "
                                                 "galaxy catalog' catalog 'pos_x' catalog field\\]."):
            del tobj.object_properties[x.property_name]

        # Delete CatalogField from Catalog first
        del cat.catalog_fields[x.property_name]

        # THEN delete Property from TargetObject => Ok
        del tobj.object_properties[x.property_name]

    def test_catalog_galactica_validity_checks(self, caplog):
        """
        Tests catalog Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        # Add tabulated values in catalog + properties in target object
        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        z = tobj.object_properties.add(ObjectProperty(property_name="pos_z"))
        cat = Catalog(target_object=tobj, name="Spiral galaxy catalog at $Z \\simeq 1.0$", description="My descr.")

        # Empty catalog nofield, no object within
        cat.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'Spiral galaxy catalog at $Z \\simeq 1.0$' catalog does not contain any "
                                            "object.")
        cat.catalog_fields.add(CatalogField(obj_prop=x, values=N.random.uniform(size=100)))

        # Catalog name too long
        cat.name = "This is a way too long catalog name for a 'Spiral galaxy catalog at $Z \\simeq 1.0$'"
        cat.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'This is a way too long catalog name for a 'Spiral galaxy catalog at $Z "
                                            "\\simeq 1.0$'' catalog name is too long for Galactica (max. 64 "
                                            "characters).")

    def test_catalog_hdf5_io(self, tmp_path):
        """
        Tests saving/loading Catalog from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        # Add tabulated values in catalog + properties in target object
        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        z = tobj.object_properties.add(ObjectProperty(property_name="pos_z"))
        cat = Catalog(target_object=tobj, name="Spiral galaxy catalog at $Z \\simeq 1.0$", description="My descr.")
        vals_x = N.random.uniform(size=20)
        cat.catalog_fields.add(CatalogField(obj_prop=x, values=vals_x))

        cat.datafiles.add(Datafile(name="My datafile"))
        cat.datafiles.add(Datafile(name="My datafile (2)"))
        res = GenericResult(name="Key result 1 !", description="My description", directory_path="/my/path/to/result")
        res.catalogs.add(cat)

        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.generic_results.add(res)

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]

        # Compare catalogs
        res_loaded = simu_loaded.generic_results[res.name]
        cat_loaded = res_loaded.catalogs[cat.name]
        assert cat == cat_loaded and cat.nobjects == cat_loaded.nobjects and cat_loaded.nobjects == 20

    def test_catalog_targobj_hdf5_io_compatibility_v1(self):
        """Backward compatibility test : tests loading a Simulation study saved with version v<0.5 where Catalog and
        TargetObject instances did not exist."""
        # Try to load a Project with no catalog/target object (v1), saved by an older version of astrophysix (v<0.5)
        study_0_4_2_path = os.path.join(os.path.dirname(__file__), "io", "backward_compat", "study_v0.4.2.h5")
        study = SimulationStudy.load_HDF5(study_0_4_2_path)
        # No catalog
        assert len(study.project.simulations[0].snapshots[0].catalogs) == 0

    def test_catalog_to_pandas(self):
        """Tests Catalog conversion into a pandas.Dataframe"""
        x = ObjectProperty(property_name="pos_x", unit=U.kpc)
        y = ObjectProperty(property_name="pos_y", unit=U.kpc)
        z = ObjectProperty(property_name="pos_z", unit=U.kpc)
        tobj = TargetObject(name="galaxy")
        tobj.object_properties.add(x)
        tobj.object_properties.add(y)
        tobj.object_properties.add(z)
        cat = Catalog(target_object=tobj, name="catalog #1")
        fx = cat.catalog_fields.add(CatalogField(obj_prop=x, values=N.random.uniform(size=200)*10.0+100.0))
        fy = cat.catalog_fields.add(CatalogField(obj_prop=y, values=N.random.uniform(size=200)*3.0+50.0))
        fz = cat.catalog_fields.add(CatalogField(obj_prop=z, values=N.random.uniform(size=200)*4.0+250.0))
        df = cat.to_pandas()
        assert df.shape == (200, 3)
        assert (fx.field_values == df[fx.object_property.display_name].values).all()
        assert (fy.field_values == df[fy.object_property.display_name].values).all()
        assert (fz.field_values == df[fz.object_property.display_name].values).all()

    def test_catalog_hdf5_standalone_export(self, tmp_path):
        """
        Tests exporting Catalog into a standalone HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        # Add tabulated values in catalog + properties/property group in target object
        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        z = tobj.object_properties.add(ObjectProperty(property_name="pos_z"))
        m = tobj.object_properties.add(ObjectProperty(property_name="mass"))
        pos_group = ObjectPropertyGroup(group_name="position")
        pos_group.group_properties.add(x)
        pos_group.group_properties.add(y)
        pos_group.group_properties.add(z)
        tobj.property_groups.add(pos_group)

        cat = Catalog(target_object=tobj, name="Spiral galaxy catalog at $Z \\simeq 1.2$",
                      description="My description of the catalog")
        vals_x = N.random.uniform(size=20) * 3.0 + 10.0
        vals_y = N.random.uniform(size=20) * 2.0 + 50.0
        vals_z = N.random.uniform(size=20) * 5.0 + 100.0
        vals_m = N.random.uniform(size=20) * 40.0 + 1000.0
        cat.catalog_fields.add(CatalogField(obj_prop=x, values=vals_x))
        cat.catalog_fields.add(CatalogField(obj_prop=y, values=vals_y))
        cat.catalog_fields.add(CatalogField(obj_prop=z, values=vals_z))
        cat.catalog_fields.add(CatalogField(obj_prop=m, values=vals_m))

        # Save catalog into a standalone HDF5 file
        h5_fpath = str(tmp_path / "catalog.h5")
        cat.save_HDF5(h5_fpath, galactica_checks=True)

        # Reread catalog HDF5 file
        h5f, h5group, close_when_done = Hdf5StudyPersistent.open_h5file(h5_fpath, mode=Hdf5StudyFileMode.READ_ONLY)
        try:
            catalog_group = h5group["CATALOG"]
            loaded_cat = Catalog.hsp_load_from_h5(catalog_group, dependency_objdict={})
            assert cat == loaded_cat
        except Exception:
            raise
        finally:
            if close_when_done and h5f is not None:
                h5f.close()


class TestCatalogDataProcessingService(object):
    def test_process_service_comparison(self):
        """
        Tests rich comparison method CatalogDataProcessingService.__eq__()
        """
        dps = CatalogDataProcessingService(service_name="my_service", data_host="my_workstation")

        # Identical data processing service
        dps2 = CatalogDataProcessingService(service_name=dps.service_name, uid=dps.uid, data_host=dps.data_host)
        assert dps == dps2

        # Compare with different catalog field bindings
        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        y = tobj.object_properties.add(ObjectProperty(property_name="pos_y"))
        z = tobj.object_properties.add(ObjectProperty(property_name="pos_z"))
        vals_x = N.random.uniform(size=20) * 3.0 + 10.0
        vals_y = N.random.uniform(size=20) * 2.0 + 50.0
        vals_z = N.random.uniform(size=20) * 5.0 + 100.0
        cf_x = CatalogField(obj_prop=x, values=vals_x)
        cf_y = CatalogField(obj_prop=y, values=vals_y)
        cf_z = CatalogField(obj_prop=z, values=vals_z)
        bx = dps.catalog_field_bindings.add(CatalogFieldBinding(catalog_field=cf_x, param_key="xcenter"))
        by = dps.catalog_field_bindings.add(CatalogFieldBinding(catalog_field=cf_y, param_key="ycenter"))
        bz = dps.catalog_field_bindings.add(CatalogFieldBinding(catalog_field=cf_z, param_key="zcenter"))
        # Catalog field binding list differ => not equals
        assert dps != dps2
        dps2.catalog_field_bindings.add(bx)
        dps2.catalog_field_bindings.add(by)
        dps2.catalog_field_bindings.add(bz)
        # Identical CatalogDataProcessingService objects
        assert dps == dps2

    def test_cannot_add_data_proc_service_in_genresult_catalogs(self):
        """
        Tests that trying to add a CatalogDataProcessingService into a GerericResult's catalog raises an error.
        """
        dps1 = CatalogDataProcessingService(service_name="slice_map", data_host="my_cluster")
        dps2 = CatalogDataProcessingService(service_name="ppv_cube", data_host="my_GPU_cluster")

        tobj = TargetObject(name="Spiral galaxy")
        sn = Snapshot(name="My snapshot")
        gres = GenericResult(name="My result")
        gres_cat = Catalog(target_object=tobj, name="Spiral galaxy catalog")
        gres_cat2 = Catalog(target_object=tobj, name="Spiral galaxy catalog 2")
        gres.catalogs.add(gres_cat2)
        gres_cat.processing_services.add(dps1)

        # Tries to add a a catalog linked to data-processing services into a GenericResult => error
        with pytest.raises(AttributeError, match="Cannot add a 'Spiral galaxy catalog' catalog with any data-processing"
                                                 " service linked to it into a GenericResult."):
            gres.catalogs.add(gres_cat)

        # Tries to add a data-processing service into a GenericResult's catalog => error
        with pytest.raises(AttributeError, match="Cannot add a data-processing service into a GenericResult's "
                                                 "catalog."):
            gres_cat2.processing_services.add(dps2)

        # With snapshots, should work. (# 1 first add catalog into snapshot, then add dps to catalog)
        cat_sn = sn.catalogs.add(Catalog(target_object=tobj, name="Spiral galaxy catalog"))
        cat_sn.processing_services.add(dps2)

        # (#2 first add dps to catalog, then add catalog into snapshot)
        cat_sn2 = Catalog(target_object=tobj, name="Spiral galaxy catalog 2")
        cat_sn2.processing_services.add(dps2)
        sn.catalogs.add(cat_sn2)

    def test_data_proc_service_hdf5_io(self, tmp_path):
        """
        Tests saving/loading CatalogDataProcessingService from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        dps1 = CatalogDataProcessingService(service_name="slice_map", data_host="my_cluster")
        dps2 = CatalogDataProcessingService(service_name="ppv_cube", data_host="my_GPU_cluster")

        tobj = TargetObject(name="Spiral galaxy")
        cat = Catalog(target_object=tobj, name="Spiral galaxy catalog")
        # Add data processing services
        dps1 = cat.processing_services.add(dps1)
        dps2 = cat.processing_services.add(dps2)

        sn = Snapshot(name="My snapshot")
        sn.catalogs.add(cat)

        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.snapshots.add(sn)

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]
        sn_loaded = simu_loaded.snapshots[sn.name]
        cat_loaded = sn_loaded.catalogs[cat.name]

        # Compare services
        dps1_loaded = cat_loaded.processing_services[dps1.hosted_service]
        dps2_loaded = cat_loaded.processing_services[dps2.hosted_service]
        assert dps1 == dps1_loaded
        assert dps2 == dps2_loaded

    def test_data_proc_service_galactica_validity_checks(self, caplog):
        """
        Tests data processing service Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        pass


class TestCatalogFieldBinding(object):
    def test_catalog_field_binding_init(self):
        """
        Tests object catalog field binding instance initialisation
        """
        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        vals_x = N.random.uniform(size=20) * 3.0 + 10.0
        cf_x = CatalogField(obj_prop=x, values=vals_x)

        # Object catalog initialisation
        fbx = CatalogFieldBinding(catalog_field=cf_x, param_key="alpha", scale=2.0, offset=0.5)

        assert fbx.field_property_name == "pos_x"
        assert fbx.catalog_field is cf_x
        assert fbx.param_key == "alpha"
        assert  fbx.scale == 2.0
        assert fbx.offset == 0.5

        # Test str() conversion
        assert str(fbx) == "'pos_x' catalog field binding"

    def test_catalog_field_binding_catfield_setting_exc(self):
        """
        Tests setting a CatalogField into a CatalogFieldBinding object
        """
        # Tests CatalogFieldBinding defined without CatalogField
        with pytest.raises(AttributeError, match="Undefined 'catalog_field' attribute in CatalogFieldBinding."):
            catf_exc = CatalogFieldBinding(param_key="alpha")  # No catalog field defined

        # Tests invalid catalog_field initialisation exception
        invalid_catfield_err = "CatalogFieldBinding 'catalog_field' attribute is not a valid CatalogField instance."
        with pytest.raises(AttributeError, match=invalid_catfield_err):
            catf_exc = CatalogFieldBinding(catalog_field={2: "abcd"}, param_key="My catalog")  # Invalid catalog field
        with pytest.raises(AttributeError, match=invalid_catfield_err):
            catf_exc = CatalogFieldBinding(-5, param_key="My catalog")  # Invalid catalog field

        # Catalog field initialisation : valid => should not raise any exception
        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        vals_x = N.random.uniform(size=20) * 3.0 + 10.0
        cf_x = CatalogField(obj_prop=x, values=vals_x)

        fb1 = CatalogFieldBinding(catalog_field=cf_x, param_key="alpha")
        fb2 = CatalogFieldBinding(cf_x, param_key="alpha")

    def test_setting_catalog_field_binding_param_key_exc(self):
        """
        Tests that a catalog field binding defined without a param_key  raises an exception
        """
        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        vals_x = N.random.uniform(size=20) * 3.0 + 10.0
        cf_x = CatalogField(obj_prop=x, values=vals_x)

        # Tests catalog field binding defined without parameter key
        with pytest.raises(AttributeError, match="CatalogFieldBinding 'param_key' attribute is not defined \(mandatory\)."):
            fb1 = CatalogFieldBinding(catalog_field=cf_x)

        fb1 = CatalogFieldBinding(catalog_field=cf_x, param_key="alpha")
        # ------------------- Tests invalid catalog field binding param_key setting exception ----------------------- #
        empty_param_key_err = "CatalogFieldBinding 'param_key' property is not a valid \(non-empty\) string."
        with pytest.raises(AttributeError, match=empty_param_key_err):
            fb1.param_key = -1

        with pytest.raises(AttributeError, match=empty_param_key_err):
            fb1.param_key = ""

        # ----------------------------------------------------------------------------------------------------------- #

        # Valid => should not raise any exception
        fb1.name = "beta"

    def test_setting_catalog_field_binding_scale_offset_exc(self):
        """
        Tests that a catalog field binding defined with an invalid scale/offset value raises an exception
        """
        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        vals_x = N.random.uniform(size=20) * 3.0 + 10.0
        cf_x = CatalogField(obj_prop=x, values=vals_x)
        pkey = "alhpa"

        # ---------------- Tests catalog field binding defined with an invalid scale/offset value -------------------- #
        invalid_scale_num_value_err = "CatalogFieldBinding 'scale' attribute is not a valid numeric value."
        with pytest.raises(AttributeError, match=invalid_scale_num_value_err):
            fb1 = CatalogFieldBinding(catalog_field=cf_x, param_key=pkey, scale=(-4, "Planet", True))
        with pytest.raises(AttributeError, match=invalid_scale_num_value_err):
            fb1 = CatalogFieldBinding(catalog_field=cf_x, param_key=pkey, scale="Black hole")
        invalid_offset_num_value_err = "CatalogFieldBinding 'offset' attribute is not a valid numeric value."
        with pytest.raises(AttributeError, match=invalid_offset_num_value_err):
            fb1 = CatalogFieldBinding(catalog_field=cf_x, param_key=pkey, offset=("A", 1.256))
        with pytest.raises(AttributeError, match=invalid_offset_num_value_err):
            fb1 = CatalogFieldBinding(catalog_field=cf_x, param_key=pkey, offset="Galaxy")
        # ------------------------------------------------------------------------------------------------------------ #

        # Valid => should not raise any exception
        fb = CatalogFieldBinding(catalog_field=cf_x, param_key=pkey, scale=1.0e6, offset=-100.0)
        for valid_num_val in [0.52, -256.2, "1.0e3", "-45.2"]:
            fb.scale = valid_num_val
            fb.offset = valid_num_val

    def test_cat_field_binding_comparison(self):
        """
        Tests rich comparison method CatalogFieldBinding.__eq__()
        """
        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        vals_x = N.random.uniform(size=20) * 3.0 + 10.0
        cf_x = CatalogField(obj_prop=x, values=vals_x)
        cfb = CatalogFieldBinding(catalog_field=cf_x, param_key="xcenter", scale=4.0, offset=-1.0)

        # Different UUID => not equals
        assert cfb != CatalogFieldBinding(cf_x, param_key=cfb.param_key, scale=cfb.scale, offset=cfb.offset)

        # Different catalog field => not equals
        assert cfb != CatalogFieldBinding(CatalogField(obj_prop=x, values=vals_x*3.0), param_key=cfb.param_key,
                                          uid=cfb.uid, scale=cfb.scale, offset=cfb.offset)

        # Different service parameter key => not equals
        assert cfb != CatalogFieldBinding(cf_x, param_key="alpha", uid=cfb.uid, scale=cfb.scale, offset=cfb.offset)

        # Different scaling/offset num. values
        assert cfb != CatalogFieldBinding(cf_x, param_key=cfb.param_key, uid=cfb.uid, scale=cfb.scale, offset=1.5)
        assert cfb != CatalogFieldBinding(cf_x, param_key=cfb.param_key, uid=cfb.uid, scale=1.2, offset=cfb.offset)

        # Identical catalog field bindings
        cfb2 = CatalogFieldBinding(cf_x, param_key=cfb.param_key, uid=cfb.uid, scale=cfb.scale, offset=cfb.offset)
        assert cfb == cfb2

    def test_cat_field_binding_hdf5_io(self, tmp_path):
        """
        Tests saving/loading CatalogFieldBinding from HDF5 file

        Parameters
        ----------
        tmp_path: temporary path  PyTest fixture
        """
        dps1 = CatalogDataProcessingService(service_name="slice_map", data_host="my_cluster")

        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x"))
        vals_x = N.random.uniform(size=20) * 3.0 + 10.0
        cat = Catalog(target_object=tobj, name="Spiral galaxy catalog")
        cf_x = cat.catalog_fields.add(CatalogField(obj_prop=x, values=vals_x))

        # Add data processing service
        fbx = dps1.catalog_field_bindings.add(CatalogFieldBinding(catalog_field=cf_x, param_key="xcenter", scale=1.0e6,
                                                                  offset=-100.0))
        dps1 = cat.processing_services.add(dps1)

        sn = Snapshot(name="My snapshot")
        sn.catalogs.add(cat)

        ramses = SimulationCode(name="Ramses 3 (MHD)", code_name="Ramses")
        simu = Simulation(simu_code=ramses, name="My simu énorme")
        simu.snapshots.add(sn)

        # Dummy project
        proj = Project(category=ProjectCategory.GalacticDynamics, project_title="My project")
        proj.simulations.add(simu)

        # Save study
        fname = str(tmp_path / "study.h5")
        study = SimulationStudy(project=proj)
        study.save_HDF5(fname)

        # Reload study
        study_loaded = SimulationStudy.load_HDF5(fname)
        simu_loaded = study_loaded.project.simulations[simu.name]
        sn_loaded = simu_loaded.snapshots[sn.name]
        cat_loaded = sn_loaded.catalogs[cat.name]
        dps1_loaded = cat_loaded.processing_services[dps1.hosted_service]

        # Compare catalog field bindings
        assert fbx == dps1_loaded.catalog_field_bindings[fbx.param_key]

    def test_cat_field_binding_galactica_validity_checks(self, caplog):
        """
        Tests catalog data processing service field binding Galactica validity checks

        Parameters
        ----------
        caplog: captured log PyTest fixture
        """
        dps1 = CatalogDataProcessingService(service_name="slice_map", data_host="my_cluster")

        tobj = TargetObject(name="Spiral galaxy")
        x = tobj.object_properties.add(ObjectProperty(property_name="pos_x", dtype=DataType.REAL))
        cl = tobj.object_properties.add(ObjectProperty(property_name="class", dtype=DataType.STRING))

        vals_x = N.random.uniform(size=20) * 3.0 + 10.0
        vals_cl = N.array(["A"]*20)
        cat = Catalog(target_object=tobj, name="Spiral galaxy catalog")
        cf_x = cat.catalog_fields.add(CatalogField(obj_prop=x, values=vals_x))

        # Set a catalog field binding on 'cf_x' catalog field associated with a 'long_serv_param_name' data processing
        # service parameter (name of the parameter is too long)
        fbx = dps1.catalog_field_bindings.add(CatalogFieldBinding(catalog_field=cf_x,
                                                                  param_key="long_serv_param_name"))
        # Add data processing service
        dps1 = cat.processing_services.add(dps1)

        # Check data-processing service parameter name length
        cat.galactica_validity_check()
        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "Data processing service parameter key 'long_serv_param_name' is too long "
                                            "for Galactica (max. 16 characters).")

        # Check catalog field associated object property type (is it numerical ?)
        cf_cl = cat.catalog_fields.add(CatalogField(obj_prop=cl, values=vals_cl))
        # Set a catalog field binding on 'cf_cl' catalog field associated to a STRING datatyped object property
        # (not a numerical property)
        fb_cl = dps1.catalog_field_bindings.add(CatalogFieldBinding(catalog_field=cf_cl, param_key="alpha"))
        cat.galactica_validity_check()

        assert caplog.record_tuples[-1] == ('astrophysix.simdm', logging.WARNING,
                                            "'class' catalog field binding is not a numerical type, it cannot be bound "
                                            "to a data processing service parameter.")


__all__ = ["TestCatalog", "TestCatalogDataProcessingService", ]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""

Numerical :obj:`Experiments` can be of two different types:

* :class:`~astrophysix.simdm.experiment.Simulation`,
* :class:`~astrophysix.simdm.experiment.PostProcessingRun`.

.. autoclass:: astrophysix.simdm.experiment.Simulation
   :members:
   :undoc-members:
   :inherited-members:
   :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5,
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX

.. autoclass:: astrophysix.simdm.experiment.PostProcessingRun
   :members:
   :undoc-members:
   :inherited-members:
   :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5,
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX

"""
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str, list, dict
import logging
import uuid
from datetime import datetime

from ..datafiles import JsonFile, AsciiFile, YamlFile
from ..datafiles.file import AssociatedFile
from ...utils.strings import Stringifiable
from ...utils.persistency import Hdf5StudyPersistent
from ...utils import DatetimeUtil

from ..protocol import SimulationCode, PostProcessingCode
from ..protocol.base import Protocol
from ..utils import ObjectList, GalacticaValidityCheckMixin
from .param_setting import ParameterSetting
from .app_algo import AppliedAlgorithm
from .resolved_physics import ResolvedPhysicalProcess
from ..results import Snapshot, GenericResult


log = logging.getLogger("astrophysix.simdm")


class Experiment(Hdf5StudyPersistent, GalacticaValidityCheckMixin):
    """
    Abstract Experiment class

    Parameters
    ----------
    name: :obj:`string`
        Experiment name (mandatory)
    alias: :obj:`string`
        Experiment alias (if defined, 16 max characters is recommended)
    description: :obj:`string`
        Long experiment description
    directory_path: :obj:`string`
        Experiment directory path
    config_file: :class:`~astrophysix.simdm.datafiles.file.JsonFile` or :class:`~astrophysix.simdm.datafiles.file.YamlFile` or :class:`~astrophysix.simdm.datafiles.file.AsciiFile`
        Experiment configuration file or None
    """
    def __init__(self, **kwargs):
        super(Experiment, self).__init__(**kwargs)

        self._alias = ""
        self._name = ""
        self._description = ""
        self._protocol = None
        self._directory_path = ""
        self._results = ObjectList(GenericResult, "name")
        self._snapshots = ObjectList(Snapshot, "name")
        self._psettings = ObjectList(ParameterSetting, "parameter_key")
        self._psettings.add_validity_check_method(self._check_valid_input_parameter)
        self._app_algos = ObjectList(AppliedAlgorithm, "algo_name")
        self._app_algos.add_validity_check_method(self._check_valid_algorithm)
        self._config_file = None

        # Simulation name
        if "name" not in kwargs:
            raise AttributeError("{cname:s} 'name' attribute is not defined (mandatory).".format(cname=self.__class__.__name__))
        self.name = kwargs["name"]

        if "alias" in kwargs:
            self.alias = kwargs["alias"]

        if "description" in kwargs:
            self.description = kwargs["description"]

        if "directory_path" in kwargs:
            self.directory_path = kwargs["directory_path"]

        if "config_file" in kwargs:
            self.configuration_file = kwargs["config_file"]

    def __eq__(self, other):
        """
        Experiment comparison method

        other: :obj:`Experiment`
            experiment to compare to
        """
        if not super(Experiment, self).__eq__(other):
            return False

        if self._name != other.name:
            return False

        if self._alias != other.alias:
            return False

        if self._description != other.description:
            return False

        if self._directory_path != other.directory_path:
            return False

        if self._psettings != other.parameter_settings:
            return False

        if self._app_algos != other.applied_algorithms:
            return False

        if self._snapshots != other.snapshots:
            return False

        if self._results != other.generic_results:
            return False

        if self._config_file != other.configuration_file:
            return False

        return True

    def _check_valid_input_parameter(self, param_setting):
        """
        Checks that a given parameter setting can be added into this experiment parameter setting list. Verifies that
        the parameter setting's input parameter belongs to the experiment protocol's input parameter list. Raises an
        AttributeError if not.

        Parameters
        ----------
        param_setting: ``astrophysix.simdm.experiment.param_setting.ParameterSetting``
            parameter setting to add
        """
        if param_setting.input_parameter not in self._protocol.input_parameters:
            err_msg = "{cname:s} '{ps!s}' does not refer to one of the input parameters of " \
                      "'{prot!s}'.".format(cname=self.__class__.__name__, ps=param_setting, prot=self._protocol)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def _can_delete_input_param(self, inp_param):
        """
        Checks if an input parameter is not linked to any experiment's parameter setting and can be safely deleted.
        Returns None if it can be deleted, otherwise returns a string.

        Parameters
        ----------
        inp_param: ``astrophysix.simdm.protocol.input_parameters.InputParameter``
            input parameter about to be deleted

        Returns
        -------
        o: str or None
        """
        for psetting in self._psettings:
            if psetting.input_parameter is inp_param:  # Reference identity, not equality ??? Should work
                return "{s!s} {ps!s}".format(s=self, ps=psetting)
        return None

    def _check_valid_algorithm(self, applied_algo):
        """
        Checks that a given applied algorithm can be added into this experiment applied algorithm list. Verifies that
        the applied algorithm's algorithm belongs to the experiment protocol's algorithm list. Raises an
        AttributeError if not.

        Parameters
        ----------
        applied_algo: ``astrophysix.simdm.experiment.app_algo.AppliedAlgorithm``
            applied algorithm to add
        """
        if applied_algo.algorithm not in self._protocol.algorithms:
            err_msg = "{cname:s} '{aa!s}' does not refer to one of the algorithms of " \
                      "'{prot!s}'.".format(cname=self.__class__.__name__, aa=applied_algo, prot=self._protocol)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def _can_delete_algo(self, algo):
        """
        Checks if an algorithm is not linked to any experiment's applied algorithm and can be safely deleted.
        Returns None if it can be deleted, otherwise returns a string.

        Parameters
        ----------
        algo: ``astrophysix.simdm.protocol.algorithm.Algorithm``
            algorithm about to be deleted

        Returns
        -------
        o: str or None
        """
        for app_algo in self._app_algos:
            if app_algo.algorithm is algo: # Reference identity, not equality ??? Should work
                return "{s!s} {aa!s}".format(s=self, aa=app_algo)
        return None

    @property
    def alias(self):
        """Experiment alias. Can be edited."""
        return self._alias

    @alias.setter
    def alias(self, new_alias):
        try:
            self._alias = Stringifiable.cast_string(new_alias)
            if len(new_alias) > 16:
                log.warning("{cname:s} 'alias' attribute is too long (max 16 characters).".format(
                    cname=self.__class__.__name__))

        except TypeError:
            err_msg = "{cname:s} 'alias' property is not a valid string".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def name(self):
        """Experiment name. Can be edited."""
        return self._name

    @name.setter
    def name(self, new_simu_name):
        try:
            self._name = Stringifiable.cast_string(new_simu_name, valid_empty=False)
        except TypeError:
            err_msg = "{cname:s} 'name' property is not a valid (non-empty) string".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def description(self):
        """Experiment description. Can be edited."""
        return self._description

    @description.setter
    def description(self, new_descr):
        try:
            self._description = Stringifiable.cast_string(new_descr)
        except TypeError:
            err_msg = "{cname:s} 'description' property is not a valid string".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def directory_path(self):
        """Experiment data directory path. Can be edited."""
        return self._directory_path

    @directory_path.setter
    def directory_path(self, new_path):
        try:
            self._directory_path = Stringifiable.cast_string(new_path)
        except TypeError:
            err_msg = "{cname:s} 'directory_path' property is not a valid string".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def applied_algorithms(self):
        """Experiment applied algorithm list (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._app_algos

    @property
    def parameter_settings(self):
        """Experiment parameter setting list (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._psettings

    @property
    def snapshots(self):
        """Experiment snapshot list (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._snapshots

    @property
    def generic_results(self):
        """Experiment generic result list (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._results

    @property
    def configuration_file(self):
        """
        Experiment configuration file (:class:`~astrophysix.simdm.datafiles.file.JsonFile`,
        :class:`~astrophysix.simdm.datafiles.file.YamlFile` or :class:`~astrophysix.simdm.datafiles.file.AsciiFile`).
        Set to None to leave undefined. Can be edited."""
        return self._config_file

    @configuration_file.setter
    def configuration_file(self, new_cfg_file):
        if new_cfg_file is not None and not isinstance(new_cfg_file, JsonFile) and not \
                isinstance(new_cfg_file, AsciiFile) and not isinstance(new_cfg_file, YamlFile):
            err_msg = "{cname:s} 'configuration_file' property is neither None nor a valid JsonFile, YamlFile or " \
                      "AsciiFile instance.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._config_file = new_cfg_file

    @classmethod
    def _protocol_class_name(cls):
        raise NotImplementedError()

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["name", "alias", "description", "directory_path", "config_file"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize an Experiment object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the Experiment into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(Experiment, self)._hsp_write(h5group, **kwargs)

        # Check that protocol is defined
        if self._protocol is None:
            err_msg = "Undefined protocol for {exp!s}".format(exp=self)
            log.error(err_msg)
            raise ValueError(err_msg)

        # Write experiment name
        self._hsp_write_attribute(h5group, ('name', self._name), **kwargs)

        # Write experiment Galactica alias, if defined
        self._hsp_write_attribute(h5group, ('galactica_alias', self._alias), **kwargs)

        # Write experiment description, if defined
        self._hsp_write_attribute(h5group, ('description', self._description), **kwargs)

        # Write experiment directory path, if defined
        self._hsp_write_attribute(h5group, ('experiment_directory', self._directory_path), **kwargs)

        # Write protocol
        if kwargs.get("from_project", False):  # Write protocol UUID
            self._hsp_write_attribute(h5group, ('protocol_uid', self._protocol.uid), **kwargs)
        else:  # Write complete protocol description (full serialization)
            self._hsp_write_object(h5group, "PROTOCOL", self._protocol, **kwargs)

        # Write configuration file into HDF5 file
        self._hsp_write_object(h5group, "config_file", self._config_file, **kwargs)

        # Write parameter settings, if any defined
        # # -- Security check : checks that parameter settings refers to a actual input parameters of the protocol - #
        # for psetting in self._psettings:
        #     if psetting.input_parameter not in self._protocol.input_parameters:
        #         err_msg = "{cname:s} '{ps!s}' does not refer to one of the input parameters of '{prot!s}'" \
        #                   ".".format(cname=self.__class__.__name__, ps=psetting, prot=self._protocol)
        #         log.error(err_msg)
        #         raise ValueError(err_msg)
        # # -------------------------------------------------------------------------------------------------------- #
        self._hsp_write_object_list(h5group, "PARAM_SETTINGS", self._psettings, "psetting_", **kwargs)

        # Write applied algorithms, if any defined
        # # ---- Security check : checks that applied algorithms refers to a actual algorithms of the protocol ----- #
        # for app_algo in self._app_algos:
        #     if app_algo.algorithm not in self._protocol.algorithms:
        #         err_msg = "{cname:s} '{aa!s}' does not refer to one of the algorithms of '{prot!s}'" \
        #                   ".".format(cname=self.__class__.__name__, aa=app_algo, prot=self._protocol)
        #         log.error(err_msg)
        #         raise ValueError(err_msg)
        # # -------------------------------------------------------------------------------------------------------- #
        self._hsp_write_object_list(h5group, "APPLIED_ALGOS", self._app_algos, "app_algo_", **kwargs)

        # Write all snapshots, if any defined
        self._hsp_write_object_list(h5group, "SNAPSHOTS", self._snapshots, "sn_", **kwargs)

        # Write generic results, if any defined
        self._hsp_write_object_list(h5group, "RESULTS", self._results, "res_", **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read an Experiment object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        exp: ``Experiment``
            Read Experiment instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(Experiment, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Try to read/find protocol
        try:
            prot_uid = uuid.UUID(cls._hsp_read_attribute(h5group, "protocol_uid", "protocol UUID",
                                                         raise_error_if_not_found=True))

            # Search for already instantiated protocol in dependency object dictionary
            if dependency_objdict is None:
                err_msg = "Cannot find any protocol already instantiated in the project."
                log.error(err_msg)
                raise IOError(err_msg)

            # Get dictionary of protocol objects of the corresponding class : simulation / post-processing codes
            pcn = cls._protocol_class_name()
            if pcn not in dependency_objdict:
                err_msg = "Cannot find any protocol of type '{n:s}'".format(n=pcn)
                log.error(err_msg)
                raise IOError(err_msg)

            # Find protocol according to its UUID
            proto_dict = dependency_objdict[pcn]
            if prot_uid not in proto_dict:
                err_msg = "Cannot find protocol of type '{n:s}' with uid {uid:s}.".format(n=pcn, uid=str(prot_uid))
                log.error(err_msg)
                raise IOError(err_msg)

            prot = proto_dict[prot_uid]
        except IOError:  # Protocol UUID not found in Experiment
            # Read protocol info from "PROTOCOL" subgroup
            prot = Protocol._hsp_read_object(h5group, "PROTOCOL", "experiment protocol",
                                             dependency_objdict=dependency_objdict)

        # Read experiment name
        exp_name = cls._hsp_read_attribute(h5group, "name", "experiment name")

        # Instantiate experiment object
        exp = cls(prot, uid=uid, name=exp_name)

        # Read experiment Galactica alias, if defined
        exp_alias = cls._hsp_read_attribute(h5group, 'galactica_alias', "experiment Galactica alias",
                                            raise_error_if_not_found=False)
        if exp_alias is not None:
            exp.alias = exp_alias

        # Read experiment description, if defined
        exp_descr = cls._hsp_read_attribute(h5group, 'description', "experiment description",
                                            raise_error_if_not_found=False)
        if exp_descr is not None:
            exp.description = exp_descr

        # Read experiment directory path, if defined
        exp_dpath = cls._hsp_read_attribute(h5group, 'experiment_directory', "experiment directory path",
                                            raise_error_if_not_found=False)
        if exp_dpath is not None:
            exp.directory_path = exp_dpath

        # Read configuration file, if defined
        if "config_file" in h5group:
            exp.configuration_file = AssociatedFile._hsp_read_object(h5group, "config_file", "configuration file")

        dod = {prot.__class__.__name__: prot}
        # Build parameter setting list and add each parameter setting into experiment
        if "PARAM_SETTINGS" in h5group:
            for psetting in ParameterSetting._hsp_read_object_list(h5group, "PARAM_SETTINGS", "psetting_",
                                                                   "experiment parameter setting",
                                                                   dependency_objdict=dod):
                exp.parameter_settings.add(psetting)

        # Build applied algorithm list and add each applied algorithm into experiment
        if "APPLIED_ALGOS" in h5group:
            for app_algo in AppliedAlgorithm._hsp_read_object_list(h5group, "APPLIED_ALGOS", "app_algo_",
                                                                   "experiment applied algorithm",
                                                                   dependency_objdict=dod):
                exp.applied_algorithms.add(app_algo)

        # Build snapshot list and add each snapshot into experiment
        if "SNAPSHOTS" in h5group:
            for sn in Snapshot._hsp_read_object_list(h5group, "SNAPSHOTS", "sn_", "experiment snapshot",
                                                     dependency_objdict=dependency_objdict):
                exp.snapshots.add(sn)

        # Build generic result list and add each generic result into experiment
        if "RESULTS" in h5group:
            for res in GenericResult._hsp_read_object_list(h5group, "RESULTS", "res_", "experiment generic result",
                                                           dependency_objdict=dependency_objdict):
                exp.generic_results.add(res)

        return exp

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: `dict`
            keyword arguments (optional)
        """
        # Check experiment alias
        if len(self.alias) == 0:
            log.warning("{p!s} Galactica alias is missing.".format(p=self))
        elif len(self._alias) > 64:
            log.warning("{p!s} Galactica alias is too long (max. 64 characters).".format(p=self))
        else:
            err_msg = self.galactica_valid_alias(self._alias)
            if err_msg is not None:
                log.warning("{p!s} Galactica alias is not valid ({m:s})".format(p=self, m=err_msg))

        # Check configuration file
        if self._config_file is not None:
            self._config_file.galactica_validity_check(**kwargs)

        # Perform Galactica validity checks on parameter setting/applied algorithm list
        self._psettings.galactica_validity_check(**kwargs)
        self._app_algos.galactica_validity_check(**kwargs)

        # Check generic result/snapshot validity + ensure generic result/snapshot name unicity in this experiment
        result_names = {}
        for gres in self._results():
            gres.galactica_validity_check()
            if len(gres.name) > 0:
                if gres.name in result_names:
                    log.warning("{r1!s} and {r2!s} results share the same name. They must be "
                                "unique.".format(r1=result_names[gres.name], r2=gres))
                else:
                    result_names[gres.name] = gres
        for sn in self._snapshots():
            sn.galactica_validity_check()
            if len(sn.name) > 0:
                if sn.name in result_names:
                    log.warning("{r1!s} and {r2!s} results share the same name. They must be "
                                "unique.".format(r1=result_names[sn.name], r2=sn))
                else:
                    result_names[sn.name] = sn


class Simulation(Experiment, Stringifiable):
    """
    Simulation (Simulation data model)

    Parameters
    ----------
    name: :obj:`string`
        Simulation name (mandatory)
    simu_code: :class:`~astrophysix.simdm.protocol.SimulationCode`
        Simulation code used for this simulation (mandatory)
    alias: :obj:`string`
        Simulation alias (if defined, 16 max characters is recommended)
    description: :obj:`string`
        Long simulation description
    directory_path: :obj:`string`
        Simulation data directory path
    execution_time: :obj:`string`
        Simulation execution time in the format '%Y-%m-%d %H:%M:%S'
    config_file: :class:`~astrophysix.simdm.datafiles.file.JsonFile` or :class:`~astrophysix.simdm.datafiles.file.YamlFile` or :class:`~astrophysix.simdm.datafiles.file.AsciiFile`
        Simulation configuration file or :obj:`None`
    """
    EXETIME_FORMAT = "%Y-%m-%d %H:%M:%S"

    def __init__(self, *args, **kwargs):
        super(Simulation, self).__init__(**kwargs)

        # Protocol (simulation code)
        if len(args) > 0:
            simulation_code = args[0]
        elif "simu_code" in kwargs:
            simulation_code = kwargs["simu_code"]
        else:
            raise AttributeError("Undefined simulation code for '{sname:s}' Simulation.".format(sname=self._name))

        if not isinstance(simulation_code, SimulationCode):
            err_msg = "Simulation 'simulation_code' attribute is not a valid SimulationCode instance."
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._protocol = simulation_code

        # Simulation execution time
        self._execution_time = None
        if "execution_time" in kwargs:
            self.execution_time = kwargs["execution_time"]

        # Post-processing runs
        self._ppruns = ObjectList(PostProcessingRun, "name")
        # Resolved physical processes
        self._resolved_physics = ObjectList(ResolvedPhysicalProcess, "process_name")
        self._resolved_physics.add_validity_check_method(self._check_valid_physics)

        # Add deletion handler to the simulation code input parameter list and algorithm list
        simulation_code.input_parameters.add_deletion_handler(self._can_delete_input_param)
        simulation_code.algorithms.add_deletion_handler(self._can_delete_algo)
        simulation_code.physical_processes.add_deletion_handler(self._can_delete_phys_proc)

    def __eq__(self, other):
        """
        Simulation comparison method

        other: :class:`~astrophysix.simdm.experiment.Simulation`
            simulation to compare to
        """
        if not super(Simulation, self).__eq__(other):
            return False

        if self._protocol != other.simulation_code:
            return False

        if self._execution_time != other.execution_time:
            return False

        if self._resolved_physics != other.resolved_physics:
            return False

        if self._ppruns != other.post_processing_runs:
            return False

        return True

    def _check_valid_physics(self, res_phys):
        """
        Checks that a given resolved physical process can be added into this simulation resolved physical process list.
        Verifies that the resolved physical process's physical process belongs to the simulation protocol's physical
        process list. Raises an AttributeError if not.

        Parameters
        ----------
        res_phys: ``astrophysix.simdm.experiment.resolved_physics.ResolvedPhysicalProcess``
            resolved physical process to add
        """
        if res_phys.physical_process not in self._protocol.physical_processes:
            err_msg = "{cname:s} '{aa!s}' does not refer to one of the physical processes of " \
                      "'{prot!s}'.".format(cname=self.__class__.__name__, aa=res_phys, prot=self._protocol)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def _can_delete_phys_proc(self, phys_proc):
        """
        Checks if a physical process is not linked to any simulation's resolved physical process and can be safely
        deleted. Returns None if it can be deleted, otherwise returns a string.

        Parameters
        ----------
        phys_proc: ``astrophysix.simdm.protocol.physics.PhysicalProcess``
            physical process about to be deleted

        Returns
        -------
        o: str or None
        """
        for res_pproc in self._resolved_physics:
            if res_pproc.physical_process is phys_proc:  # Reference identity, not equality ??? Should work
                return "{s!s} {rpp!s}".format(s=self, rpp=res_pproc)
        return None

    @property
    def simulation_code(self):
        """
        :class:`~astrophysix.simdm.protocol.SimulationCode` used to run this simulation. Cannot be changed after
        simulation initialisation.
        """
        return self._protocol

    @property
    def post_processing_runs(self):
        """Simulation associated post-processing run list (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._ppruns

    @classmethod
    def _protocol_class_name(cls):
        return SimulationCode.__name__

    @property
    def execution_time(self):
        """Simulation execution date/time. Can be edited.

        Example
        -------
        >>> simu = Simulation(simu_code=gadget4, name="Maxi Cosmic", execution_time="2020-09-10 14:25:48")
        >>> simu.execution_time = '2020-09-28 18:45:24'
        """
        return self._execution_time

    @execution_time.setter
    def execution_time(self, new_time_value):
        try:
            t_str = Stringifiable.cast_string(new_time_value, valid_empty=False)
            t = datetime.strptime(t_str, self.EXETIME_FORMAT)
            self._execution_time = t_str
        except (TypeError, ValueError):
            err_msg = "{cn:s} 'execution_time' property is not a valid datetime " \
                      "string.".format(cn=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def execution_time_as_utc_datetime(self):
        """UTC execution time of the simulation (timezone aware)"""
        return DatetimeUtil.utc_from_string(self._execution_time, self.EXETIME_FORMAT)

    @property
    def resolved_physics(self):
        """Simulation resolved physical process list (:class:`~astrophysix.simdm.utils.ObjectList`)."""
        return self._resolved_physics

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        experiment_valid_attrs = super(Simulation, cls)._hsp_valid_attributes()
        experiment_valid_attrs += ["simu_code", "execution_time"]
        return experiment_valid_attrs

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a Simulation object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the Simulation into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # If necessary, call callback function with experiment name
        self._hsp_write_callback(str(self), **kwargs)

        # Call to parent class _hsp_write()
        super(Simulation, self)._hsp_write(h5group, **kwargs)

        # Simulation execution time
        self._hsp_write_attribute(h5group, ('execution_time', self._execution_time), **kwargs)

        # Write simulation resolved physics, if any defined
        # # -- Security check : checks that resolved physics refers to an actual physical process of the protocol -- #
        # for res_phys in self._resolved_physics:
        #     if res_phys.physical_process not in self._protocol.physical_processes:
        #         err_msg = "{cname:s} '{rp!s}' does not refer to one of the physical processes of '{prot!s}'" \
        #                   ".".format(cname=self.__class__.__name__, rp=res_phys, prot=self._protocol)
        #         log.error(err_msg)
        #         raise ValueError(err_msg)
        # # -------------------------------------------------------------------------------------------------------- #
        self._hsp_write_object_list(h5group, "RESOLVED_PHYSICS", self._resolved_physics, "res_phys_", **kwargs)

        # Write simulation post-processing runs, if any defined
        self._hsp_write_object_list(h5group, "POST-PRO_RUNS", self._ppruns, "pprun_", **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a Simulation object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        simu: ``Simulation``
            Read Simulation instance
        """
        # Handle different versions here

        # Read base experiment parameters
        simu = super(Simulation, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read simulation execution time, if defined
        exec_time = cls._hsp_read_attribute(h5group, 'execution_time', "simulation execution time",
                                            raise_error_if_not_found=False)
        if exec_time is not None:
            simu.execution_time = exec_time

        # Build resolved physics list and add each resolved physics into simulation, if any defined
        if "RESOLVED_PHYSICS" in h5group:
            dod = {simu.simulation_code.__class__.__name__: simu.simulation_code}
            for res_phys in ResolvedPhysicalProcess._hsp_read_object_list(h5group, "RESOLVED_PHYSICS", "res_phys_",
                                                                          "simulation resolved physical process",
                                                                          dependency_objdict=dod):
                simu.resolved_physics.add(res_phys)

        # Build post-processing run list and add each post-processing run into simulation, if any defined
        if "POST-PRO_RUNS" in h5group:
            for pprun in PostProcessingRun._hsp_read_object_list(h5group, "POST-PRO_RUNS", "pprun_",
                                                                 "simulation post-processing run",
                                                                 dependency_objdict=dependency_objdict):
                simu.post_processing_runs.add(pprun)

        return simu

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: `dict`
            keyword arguments (optional)
        """
        super(Simulation, self).galactica_validity_check(**kwargs)

        # Check that simulation execution time is defined
        if self.execution_time is None:
            log.warning("{p!s} Galactica execution time is not defined.".format(p=self))

        # Perform Galactica validity checks on resolved physical process + post-processing runs list
        self._resolved_physics.galactica_validity_check(**kwargs)
        self._ppruns.galactica_validity_check(**kwargs)

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{simname:s}' simulation".format(simname=self._name)


class PostProcessingRun(Experiment, Stringifiable):
    """
    Post-processing run (Simulation data model)

    Parameters
    ----------
    name: :obj:`string`
        post-processing run name (mandatory)
    ppcode: :class:`~astrophysix.simdm.protocol.PostProcessingCode`
        post-processing code used for this post-processing run (mandatory)
    alias: :obj:`string`
        Post-processing run alias (if defined, 16 max characters is recommended)
    description: :obj:`string`
        Long post-processing run description
    config_file: :class:`~astrophysix.simdm.datafiles.file.JsonFile` or :class:`~astrophysix.simdm.datafiles.file.YamlFile` or :class:`~astrophysix.simdm.datafiles.file.AsciiFile`
        Post-processing run configuration file or None

    """
    def __init__(self, *args, **kwargs):
        super(PostProcessingRun, self).__init__(**kwargs)

        # Protocol (simulation code)
        if len(args) > 0:
            postpro_code = args[0]
        elif "ppcode" in kwargs:
            postpro_code = kwargs["ppcode"]
        else:
            raise AttributeError("Undefined post-processing code for '{ppname:s}' Post-processing run.".format(ppname=self._name))

        if not isinstance(postpro_code, PostProcessingCode):
            err_msg = "PostProcessingRun 'postpro_code' attribute is not a valid PostProcessingCode instance."
            log.error(err_msg)
            raise AttributeError(err_msg)

        self._protocol = postpro_code

        # Add deletion handler to the post-processing code input parameter list and algorithm list
        postpro_code.input_parameters.add_deletion_handler(self._can_delete_input_param)
        postpro_code.algorithms.add_deletion_handler(self._can_delete_algo)

    def __eq__(self, other):
        """
        PostProcessingRun comparison method

        other: :class:`~astrophysix.simdm.protocol.PostProcessingRun`
            post-processing run to compare to
        """
        if not super(PostProcessingRun, self).__eq__(other):
            return False

        if self._protocol != other.postpro_code:
            return False

        return True

    @property
    def postpro_code(self):
        """:class:`~astrophysix.simdm.protocol.PostProcessingCode` used to run this post-processing run. Cannot be
        changed after post-processing run initialisation."""
        return self._protocol

    @classmethod
    def _protocol_class_name(cls):
        return PostProcessingCode.__name__

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        experiment_valid_attrs = super(PostProcessingRun, cls)._hsp_valid_attributes()
        experiment_valid_attrs.append("ppcode")
        return experiment_valid_attrs

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a PostProcessingRun object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the PostProcessingRun into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # If necessary, call callback function with post-processing run name
        self._hsp_write_callback(str(self), **kwargs)

        # Call to parent class _hsp_write()
        super(PostProcessingRun, self)._hsp_write(h5group, **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a PostProcessingRun object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        pprun: :class:`~astrophysix.simdm.experiment.PostProcessingRun`
            Read PostProcessingRun instance
        """
        # Handle different versions here

        # Read base experiment parameters
        pprun = super(PostProcessingRun, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        return pprun

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{pprun_name:s}' post-processing run".format(pprun_name=self._name)


__all__ = ["Simulation", "PostProcessingRun", "Experiment"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""
@startwbs

+ Simulation study
 +_ My Project
  + Target Objects
   +_ 'Spiral galaxy' target object
   +_ 'Elliptic galaxy' target object
    + Object properties
     +_ 'pos_x' property
     +_ 'pos_y' property
     +_ 'pos_z' property
     +_ 'mass' property

    + Object property groups
     +_ 'Position' property group
      +_ 'pos_x' property
      +_ 'pos_y' property
      +_ 'pos_z' property

  +_ Cosmological simulation
   + Generic results
    +_ Generic Result #1
     + Catalogs
      +_ [Spiral galaxy] catalog #1
      +_ [Elliptic galaxy] catalog #2
       + Tabulated data
        +_ 'pos_x' value series
        +_ 'pos_y' value series
        +_ 'pos_z' values series
        +_ 'mass' value series

@endwbs

@startuml

Result "1" *-- "many" Catalog : contains
Catalog "1" *-- "many" CatalogField : contains
TargetObject "1" *-- "many" ObjectProperty : contains
TargetObject "1" *-- "many" ObjectPropertyGroup : contains
ObjectPropertyGroup "1" *-- "many" ObjectProperty : contains
CatalogField "1" *-- "1" ObjectProperty : refers to

class Result {
    +catalogs : ObjectList

}

Catalog "1" *-- "1" TargetObject : refers to

class Catalog {
    +nobjects : int
    +Catalog_fields : ObjectList
}

class CatalogField {
    +field_values : 1D numpy.ndarray
    +object_property: ObjectProperty
}

class TargetObject {
    +name : string
    +object_properties : ObjectList
    +property_groups : ObjectList
}

class ObjectProperty {
    +property_name : string
}

class ObjectPropertyGroup {
    +group_name : string
}

@enduml

"""
from .field import CatalogField  # Import before Catalog to avoid circular import errors
from .catalog import Catalog
from .targobj import TargetObject, ObjectProperty, ObjectPropertyGroup, PropertySortFlag, PropertyFilterFlag

__doc__ = """

Target object and their properties
----------------------------------

*New in version 0.5.0*

.. autoclass:: astrophysix.simdm.catalogs.targobj.TargetObject
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5, 
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX


.. autoclass:: astrophysix.simdm.catalogs.targobj.ObjectProperty
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5, 
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX


.. autoclass:: astrophysix.simdm.catalogs.targobj.ObjectPropertyGroup
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5, 
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX


.. autoclass:: astrophysix.simdm.catalogs.targobj.PropertySortFlag
   :members:
   :undoc-members:


.. autoclass:: astrophysix.simdm.catalogs.targobj.PropertyFilterFlag
   :members:
   :undoc-members:

Object catalogs and catalog fields
----------------------------------

*New in version 0.5.0*

.. autoclass:: astrophysix.simdm.catalogs.catalog.Catalog
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5, 
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX


.. autoclass:: astrophysix.simdm.catalogs.field.CatalogField
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5, 
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX

"""

__all__ = ["Catalog", "CatalogField", "TargetObject", "ObjectProperty", "ObjectPropertyGroup", "PropertySortFlag",
           "PropertyFilterFlag"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""
@startwbs

+ Simulation study
 + Project
  + Simulation codes
   +_ 'RAMSES 3.1 (MHD)' simulation code
    + Input parameters
    + Algorithms
    + Physical processes
   +_ 'AREPO' simulation code
    + Input parameters
    + Algorithms
    + Physical processes
  + Post-processing codes
   +_ 'RADMC-3D v2.0' post-processing code
    + Input parameters
    + Algorithms

@endwbs

"""
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str, list, dict
import logging
from ..utils import ObjectList, GalacticaValidityCheckMixin
from astrophysix.utils.strings import Stringifiable
from astrophysix.utils.persistency import Hdf5StudyPersistent
from .input_parameters import InputParameter
from .algorithm import Algorithm
from .physics import PhysicalProcess


__doc__ = """

:obj:`Protocol` is the generic term for a software tool used to conduct a numerical :obj:`Experiment`. There are two
different types of protocols :

* :class:`~astrophysix.simdm.protocol.SimulationCode`,
* :class:`~astrophysix.simdm.protocol.PostProcessingCode`.

.. autoclass:: astrophysix.simdm.protocol.SimulationCode
   :members:
   :undoc-members:
   :inherited-members:
   :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5, 
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX

.. autoclass:: astrophysix.simdm.protocol.PostProcessingCode
   :members:
   :undoc-members:
   :inherited-members:
   :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5, 
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX

"""

log = logging.getLogger("astrophysix.simdm")


class Protocol(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Abstract Protocol class

    Parameters
    ----------
    name: ``string``
        protocol name (mandatory)
    code_name: ``string``
        code name (mandatory)
    alias: ``string``
        protocol alias
    url: ``string``
        reference URL
    code_version: ``string``
        code version
    description: ``string``
        protocol description
    """
    def __init__(self, **kwargs):
        super(Protocol, self).__init__(**kwargs)

        self._name = ""
        self._alias = ""
        self._url = ""
        self._code_name = ""
        self._code_version = ""
        self._description = ""

        self._inpparam_list = ObjectList(InputParameter, "name")
        self._algo_list = ObjectList(Algorithm, "name")

        if "name" not in kwargs:
            raise AttributeError("{cname:s} 'name' attribute is not defined (mandatory).".format(cname=self.__class__.__name__))
        self.name = kwargs["name"]

        if "code_name" not in kwargs:
            raise AttributeError("{cname:s} 'code_name' attribute is not defined (mandatory).".format(cname=self.__class__.__name__))
        self.code_name = kwargs["code_name"]

        if "alias" in kwargs:
            self.alias = kwargs["alias"]

        if "url" in kwargs:
            self.url = kwargs["url"]

        if "code_version" in kwargs:
            self.code_version = kwargs["code_version"]

        if "description" in kwargs:
            self.description = kwargs["description"]

    def __eq__(self, other):
        """
        Protocol comparison method

        other: :obj:`Protocol`
            Protocol to compare to
        """
        if not super(Protocol, self).__eq__(other):
            return False

        if self._name != other.name:
            return False

        if self._code_name != other.code_name:
            return False

        if self._alias != other.alias:
            return False

        if self._url != other.url:
            return False

        if self._code_version != other.code_version:
            return False

        if self._description != other.description:
            return False

        if self._inpparam_list != other.input_parameters:
            return False

        if self._algo_list != other.algorithms:
            return False

        return True

    @property
    def alias(self):
        """Protocol alias"""
        return self._alias

    @alias.setter
    def alias(self, new_alias):
        try:
            self._alias = Stringifiable.cast_string(new_alias)
        except TypeError:
            raise AttributeError("{cname:s} 'alias' property is not a valid "
                                 "string".format(cname=self.__class__.__name__))

    @property
    def name(self):
        """Protocol name"""
        return self._name

    @name.setter
    def name(self, new_protocol_name):
        try:
            self._name = Stringifiable.cast_string(new_protocol_name, valid_empty=False)
        except TypeError:
            raise AttributeError("{cname:s} 'name' property is not a valid (non empty) "
                                 "string.".format(cname=self.__class__.__name__))

    @property
    def code_name(self):
        """Protocol code name"""
        return self._code_name

    @code_name.setter
    def code_name(self, new_code_name):
        try:
            self._code_name = Stringifiable.cast_string(new_code_name, valid_empty=False)
        except TypeError:
            raise AttributeError("{cname:s} 'code_name' property is not a valid (non empty) "
                                 "string.".format(cname=self.__class__.__name__))

    @property
    def code_version(self):
        """Protocol code version"""
        return self._code_version

    @code_version.setter
    def code_version(self, new_code_version):
        try:
            self._code_version = Stringifiable.cast_string(new_code_version)
        except TypeError:
            raise AttributeError("{cname:s} 'code_version' property is not a valid "
                                 "string.".format(cname=self.__class__.__name__))

    @property
    def url(self):
        """Protocol code url"""
        return self._url

    @url.setter
    def url(self, new_url):
        try:
            self._url = Stringifiable.cast_string(new_url)
        except TypeError:
            raise AttributeError("{cname:s} 'url' property is not a valid "
                                 "string.".format(cname=self.__class__.__name__))

    @property
    def description(self):
        """Protocol description"""
        return self._description

    @description.setter
    def description(self, new_descr):
        try:
            self._description = Stringifiable.cast_string(new_descr)
        except TypeError:
            raise AttributeError("{cname:s} 'description' property is not a valid string.".format(cname=self.__class__.__name__))

    @property
    def input_parameters(self):
        """Protocol :class:`~astrophysix.simdm.protocol.InputParameter` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._inpparam_list

    @property
    def algorithms(self):
        """Protocol :class:`~astrophysix.simdm.protocol.Algorithm` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._algo_list

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["name", "code_name", "alias", "url", "code_version", "description"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a Protocol object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the Protocol into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(Protocol, self)._hsp_write(h5group, **kwargs)

        # Write protocol name
        self._hsp_write_attribute(h5group, ('name', self._name), **kwargs)

        # Write protocol code name
        self._hsp_write_attribute(h5group, ('code_name', self._code_name), **kwargs)

        # Write protocol code version, if defined
        self._hsp_write_attribute(h5group, ('code_version', self._code_version), **kwargs)

        # Write protocol Galactica alias, if defined
        self._hsp_write_attribute(h5group, ('galactica_alias', self._alias), **kwargs)

        # Write protocol URL, if defined
        self._hsp_write_attribute(h5group, ('url', self._url), **kwargs)

        # Write protocol description, if defined
        self._hsp_write_attribute(h5group, ('description', self._description), **kwargs)

        # Write input parameters, if any defined
        self._hsp_write_object_list(h5group, "INPUT_PARAMS", self._inpparam_list, "inpparam_", **kwargs)

        # Write algorithms, if any defined
        self._hsp_write_object_list(h5group, "ALGORITHMS", self._algo_list, "algo_", **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a Protocol object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        exp: ``Protocol``
            Read Protocol instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(Protocol, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read protocol name
        name = cls._hsp_read_attribute(h5group, 'name', "protocol name")

        # Read protocol code name
        code_name = cls._hsp_read_attribute(h5group, 'code_name', "protocol code name")

        # Create protocol object
        prot = cls(uid=uid, name=name, code_name=code_name)

        # Read code version, if defined
        prot_codvers = cls._hsp_read_attribute(h5group, "code_version", "protocol code version",
                                               raise_error_if_not_found=False)
        if prot_codvers is not None:
            prot.code_version = prot_codvers

        # Read protocol Galactica alias, if defined
        palias = cls._hsp_read_attribute(h5group, 'galactica_alias', "protocol Galactica alias",
                                         raise_error_if_not_found=False)
        if palias is not None:
            prot.alias = palias

        # Read protocol URL, if defined
        prot_url = cls._hsp_read_attribute(h5group, 'url', "protocol URL", raise_error_if_not_found=False)
        if prot_url is not None:
            prot.url = prot_url

        # Read protocol description, if defined
        prot_descr = cls._hsp_read_attribute(h5group, 'description', "protocol description",
                                             raise_error_if_not_found=False)
        if prot_descr is not None:
            prot.description = prot_descr

        # Build input parameter list and add each input parameter into protocol, if any define
        if "INPUT_PARAMS" in h5group:
            for inpparam in InputParameter._hsp_read_object_list(h5group, "INPUT_PARAMS", "inpparam_",
                                                                 "protocol input parameter",
                                                                 dependency_objdict=dependency_objdict):
                prot.input_parameters.add(inpparam)

        # Build algorithm list and add each algorithm into protocol, if any defined
        if "ALGORITHMS" in h5group:
            for algo in Algorithm._hsp_read_object_list(h5group, "ALGORITHMS", "algo_", "protocol algorithm",
                                                        dependency_objdict=dependency_objdict):
                prot.algorithms.add(algo)

        return prot

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Check protocol alias
        if len(self._alias) == 0:
            log.warning("{p!s} Galactica protocol alias is missing.".format(p=self))
        elif len(self._alias) > 16:
            log.warning("{p!s} Galactica protocol alias is too long (max. 16 characters).".format(p=self))
        else:
            err_msg = self.galactica_valid_alias(self._alias)
            if err_msg is not None:
                log.warning("{p!s} Galactica protocol alias is not valid ({m:s})".format(p=self, m=err_msg))

        # Check protocol name
        if len(self._name) == 0:
            log.warning("{p!s} Galactica protocol name is missing.".format(p=self))
        elif len(self._name) > 128:
            log.warning("{p!s} Galactica protocol name is too long (max. 128 characters).".format(p=self))

        # Check protocol code name
        if len(self._code_name) == 0:
            log.warning("{p!s} Galactica protocol code name is missing.".format(p=self))
        elif len(self._code_name) > 32:
            log.warning("{p!s} Galactica protocol code name is too long (max. 32 characters).".format(p=self))

        # Check protocol name
        if len(self._code_version) == 0:
            log.warning("{p!s} Galactica protocol code version is missing.".format(p=self))
        elif len(self._code_version) > 8:
            log.warning("{p!s} Galactica protocol code version is too long (max. 8 characters).".format(p=self))

        # Perform Galactica validity checks on input parameter/algorithm list
        self._inpparam_list.galactica_validity_check(**kwargs)
        self._algo_list.galactica_validity_check(**kwargs)

    def __unicode__(self):
        """
        String representation of the instance
        """
        if len(self._name) > 0:
            s = "[{name:s}]".format(name=self._name)
        elif len(self._alias) > 0:
            s = "[{alias:s}]".format(alias=self._alias)
        else:
            s = ""
        return s


class PostProcessingCode(Protocol):
    """
    Post-processing code

    Parameters
    ----------
    name: :obj:`string`
        name (mandatory)
    code_name: `:obj:`string`
        base code name (mandatory)
    alias: :obj:`string`
        code alias
    url: :obj:`string`
        reference URL
    code_version: :obj:`string`
        code version
    description: :obj:`string`
        code description
    """
    def __init__(self, **kwargs):
        super(PostProcessingCode, self).__init__(**kwargs)

    def __unicode__(self):
        """
        String representation of the instance
        """
        s = super(PostProcessingCode, self).__unicode__()
        if len(s) == 0:
            return "Post-processing code"
        else:
            s += " post-processing code"
            return s


class SimulationCode(Protocol):
    """
    Simulation code

    Parameters
    ----------
    name: :obj:`string`
        name (mandatory)
    code_name: :obj:`string`
        base code name (mandatory)
    alias: :obj:`string`
        code alias
    url: :obj:`string`
        reference URL
    code_version: :obj:`string`
        code version
    description: :obj:`string`
        code description
    """
    def __init__(self, **kwargs):
        super(SimulationCode, self).__init__(**kwargs)
        self._phys_procs_list = ObjectList(PhysicalProcess, "name")

    def __eq__(self, other):
        """
        SimulationCode comparison method

        other: :class:`~astrophysix.simdm.protocool.SimulationCode`
            simulation code to compare to
        """
        if not super(SimulationCode, self).__eq__(other):
            return False

        if self._phys_procs_list != other.physical_processes:
            return False

        return True

    @property
    def physical_processes(self):
        """Simulation code :class:`~astrophysix.simdm.protocol.PhysicalProcess` list
        (:class:`~astrophysix.simdm.utils.ObjectList`)"""
        return self._phys_procs_list

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a SimulationCode object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the SimulationCode into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write()
        super(SimulationCode, self)._hsp_write(h5group, **kwargs)

        # Write physical processes, if any defined
        self._hsp_write_object_list(h5group, "PHYS_PROCS", self._phys_procs_list, "phys_proc_", **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read an SimulationCode object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        exp: ``SimulationCode``
            Read SimulationCode instance
        """
        # Handle different versions here

        # Read base Protocol object
        prot = super(SimulationCode, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Build applied physical process list and add each physical process into simulation code, if any defined
        if "PHYS_PROCS" in h5group:
            for phys_proc in PhysicalProcess._hsp_read_object_list(h5group, "PHYS_PROCS", "phys_proc_",
                                                                   "simulation code physical process",
                                                                   dependency_objdict=dependency_objdict):
                prot.physical_processes.add(phys_proc)

        return prot

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        super(SimulationCode, self).galactica_validity_check(**kwargs)

        # Perform Galactica validity checks on physical process list
        self._phys_procs_list.galactica_validity_check(**kwargs)

    def __unicode__(self):
        """
        String representation of the instance
        """
        s = super(SimulationCode, self).__unicode__()
        if len(s) == 0:
            return "Simulation code"
        else:
            s += " simulation code"
            return s


__all__ = ["SimulationCode", "PostProcessingCode", "Protocol"]



# -*- coding: utf-8 -*-
#   This file is part of Horus.
#
#   Horus is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   Horus is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with Horus.  If not, see <http://www.gnu.org/licenses/>.
"""
.. autoclass:: astrophysix.simdm.experiment.ResolvedPhysicalProcess
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5,
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX
"""
from __future__ import absolute_import, unicode_literals
from future.builtins import str
import logging
import uuid

from astrophysix.utils.persistency import Hdf5StudyPersistent
from astrophysix.utils.strings import Stringifiable

from ..protocol import PhysicalProcess
from ..utils import GalacticaValidityCheckMixin

log = logging.getLogger("astrophysix.simdm")


class ResolvedPhysicalProcess(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Simulation resolved physical process class

    Parameters
    ----------
    physics: :class:`~astrophysix.simdm.protocol.PhysicalProcess`
        simulation code's :class:`~astrophysix.simdm.protocol.PhysicalProcess` instance (mandatory)
    details: :obj:`string`
        resolved physical process implementation details
    """
    def __init__(self, **kwargs):
        super(ResolvedPhysicalProcess, self).__init__(**kwargs)

        # Simulation associated physical process
        if "physics" not in kwargs:
            raise AttributeError("ResolvedPhysicalProcess 'physics' attribute is not defined (mandatory).")
        phys = kwargs["physics"]
        if not isinstance(phys, PhysicalProcess):
            raise AttributeError("ResolvedPhysicalProcess 'physics' attribute is not a valid PhysicalProcess object.")
        self._physics = phys

        # Resolved physical process implementation details
        self._details = ""
        if "details" in kwargs:
            self.implementation_details = kwargs["details"]

    def __eq__(self, other):
        """
        ResolvedPhysicalProcess comparison method

        other: :class:`~astrophysix.simdm.experiment.ResolvedPhysicalProcess`
            resolved physical process to compare to
        """
        if not super(ResolvedPhysicalProcess, self).__eq__(other):
            return False

        if self._physics != other.physical_process:
            return False

        if self._details != other.implementation_details:
            return False

        return True

    @property
    def physical_process(self):
        """Simulation code's :class:`~astrophysix.simdm.protocol.PhysicalProcess`. Cannot be edited after instance
        initialisation"""
        return self._physics

    @property
    def process_name(self):
        """Simulation code's :class:`~astrophysix.simdm.protocol.PhysicalProcess` name. Cannot be edited."""
        return self._physics.name

    @property
    def implementation_details(self):
        """Resolved physical process implementation details. Editable."""
        return self._details

    @implementation_details.setter
    def implementation_details(self, new_details):
        try:
            self._details = Stringifiable.cast_string(new_details)
        except TypeError:
            err_msg = "{cname:s} 'implementation_details' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["physics", "details"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a ResolvedPhysicalProcess object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the ResolvedPhysicalProcess into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(ResolvedPhysicalProcess, self)._hsp_write(h5group, **kwargs)

        # Write algorithm UUID
        self._hsp_write_attribute(h5group, ('physics_uid', self._physics.uid), **kwargs)

        # Write applied algorithm implementation details, if defined
        self._hsp_write_attribute(h5group, ('details', self._details), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a ResolvedPhysicalProcess object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        res_phys: ``ResolvedPhysicalProcess``
            Read ResolvedPhysicalProcess instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(ResolvedPhysicalProcess, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read algorithm UUID
        phys_uid = uuid.UUID(cls._hsp_read_attribute(h5group, 'physics_uid', "physical process UUID"))
        protocol = list(dependency_objdict.values())[0]
        found_pproc = protocol.physical_processes.find_by_uid(phys_uid)
        if found_pproc is None:
            err_msg = "Unknown physical process UUID '{puid!s}' in the {proto!s}.".format(puid=phys_uid, proto=protocol)
            log.error(err_msg)
            raise IOError(err_msg)

        # Create resolved physical process object
        res_phys = cls(uid=uid, physics=found_pproc)

        # Read resolved physical process implementation details, if defined
        phys_details = cls._hsp_read_attribute(h5group, 'details', "physical process implementation details",
                                               raise_error_if_not_found=False)
        if phys_details is not None:
            res_phys.implementation_details = phys_details

        return res_phys

    def __unicode__(self):
        """
        String representation of the instance
        """
        s = "[{resphys_name:s}] resolved physical process".format(resphys_name=self._physics.name)
        return s


__all__ = ["ResolvedPhysicalProcess"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility

import uuid
import numpy as N
import pandas as pd
from future.builtins import str, list, int
import logging

from .targobj import ObjectProperty, PropertyFilterFlag
from ..utils import GalacticaValidityCheckMixin, DataType
from astrophysix.utils import Hdf5StudyPersistent, Stringifiable, NumpyUtil


log = logging.getLogger("astrophysix.simdm")


class CatalogField(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    def __init__(self, *args, **kwargs):
        """
        Object catalog field class

        Parameters
        ----------
        obj_prop: :class:`~astrophysix.simdm.catalogs.targobj.ObjectProperty`
            target object property (mandatory)
        values: 1D :class:`numpy.ndarray`
            catalog field value series/1D array (mandatory)
        """
        super(CatalogField, self).__init__(**kwargs)

        # Target object property
        self._obj_property = None
        if len(args) > 0:
            obj_prop = args[0]
        elif "obj_prop" in kwargs:
            obj_prop = kwargs["obj_prop"]
        else:
            raise AttributeError("Undefined 'obj_prop' object property attribute in "
                                 "{cname:s}.".format(cname=self.__class__.__name__))

        if not isinstance(obj_prop, ObjectProperty):
            err_msg = "{cname:s} 'obj_prop' attribute is not a valid ObjectProperty " \
                      "instance.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._obj_property = obj_prop

        # Field values for all catalog items
        self._val_series = None
        self._values_md5sum = None
        self._values_min = None
        self._values_max = None
        self._values_mean = None
        self._values_std = None
        self._nobj = 0

        if not kwargs.get("hdf5_init", False):
            if "values" not in kwargs:
                raise AttributeError("Undefined 'values' property in {cname:s}.".format(cname=self.__class__.__name__))
            self.field_values = kwargs["values"]

    def __eq__(self, other):
        """
        CatalogField comparison method

        other: :class:`~astrophysix.simdm.catalogs.field.CatalogField`
            catalog field object to compare to
        """
        if not super(CatalogField, self).__eq__(other):
            return False

        # Compare target object property
        if self._obj_property != other.object_property:
            return False

        # Compare catalog field value MD5 sums
        if self._values_md5sum != other._values_md5sum:
            return False

        return True

    @property
    def property_name(self):
        """Associated target object property name"""
        return self._obj_property.property_name

    @property
    def object_property(self):
        """Associated target object property (:class:`~astrophysix.simdm.catalogs.targobj.ObjectProperty`)"""
        return self._obj_property

    @property
    def field_values(self):
        """
        Catalog field values. Can be edited.

        Returns
        -------
        vals: :obj:`1D numpy.ndarray`
        """
        self._hsp_lazy_read()
        return self._val_series

    @field_values.setter
    def field_values(self, new_vals):
        """
        Set catalog field values arrays. The new array must be of size strictly positive and leave the its size
        unchanged.

        Parameters
        ----------
        new_vals: :obj:`1D numpy.ndarray`
            catalog field value 1D array

        Raises
        ------
        Aerr: :class:`AttributeError`
            an error is raised if the new array does not contain any value or if its size is different from the previous
            value array.

        Example
        -------

            >>> pass
        """
        # Record new catalog field value array
        try:
            NumpyUtil.check_is_array(new_vals, ndim=1)
        except AttributeError:
            err_msg = "'values' {cname:s} attribute must be a 1-dimensional " \
                      "array.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

        # Check data type
        kind = NumpyUtil.check_dtype(new_vals)
        if kind is None:
            err_msg = "'values' {cname:s} attribute must contain boolean/float/integer/string " \
                      "data.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

        if self._nobj == 0:  # CatalogField initialisation
            # Array size => number of items in catalog
            if new_vals.size == 0:  # No value in catalog field !
                err_msg = "{cname:s} 'values' array need at least 1 value.".format(cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)
            self._nobj = new_vals.size
        else:  # Field values were previously set. This is an update.
            # Check that array size (number of catalog items) is unchanged
            if self._nobj != new_vals.size:
                err_msg = "The number of items in this {cname:s} is {no:d} and cannot be " \
                          "changed.".format(no=self._nobj, cname=self.__class__.__name__)
                log.error(err_msg)
                raise AttributeError(err_msg)
        self._val_series = N.ascontiguousarray(new_vals.copy())  # Make a copy so that external modification won't impact the catalog field
        if kind == "s":  # String data array
            # Force string dataset encoding as bytes (h5py compatibility, numpy string "U<8" is not supported)
            self._val_series = self._val_series.astype("S")

            self._values_min = 0.0
            self._values_max = 0.0
            self._values_std = 0.0
            self._values_mean = 0.0
        else:
            self._values_max = self._val_series.max()
            self._values_min = self._val_series.min()
            self._values_mean = self._val_series.mean()
            self._values_std = self._val_series.std()

        self._values_md5sum = NumpyUtil.md5sum(new_vals)

        # Flag the CatalogField instance as 'loaded in memory' to avoid reading data from HDF5 study file in the future
        self._hsp_set_lazy_read()

    @property
    def field_values_md5sum(self):
        return self._values_md5sum

    @property
    def field_value_stats(self):
        """Returns (min., max., mean, std) tuple for this field value array"""
        return self._values_min, self._values_max, self._values_mean, self._values_std

    @property
    def nobjects(self):
        """Returns the number of objects in this catalog field => size of the field value 1D array"""
        return self._nobj

    def to_pandas(self):
        """
        Convert a CatalogField into a :class:`pandas.Series` object
        """
        return pd.Series(data=self.field_values, name=self.object_property.display_name, index=range(1, self._nobj+1))

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["obj_prop", "hdf5_init", "values"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a CatalogField object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the CatalogField into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(CatalogField, self)._hsp_write(h5group, **kwargs)

        # Write object property UUID
        self._hsp_write_attribute(h5group, ('object_property_uid', self._obj_property.uid), **kwargs)

        # Write catalog field values + stats attributes
        values_attribute_name = "_val_series"  # => self._val_series
        write_values = self._hsp_write_dataset(h5group, ("field_values", values_attribute_name),
                                               md5sum_params=("field_values_md5sum", self._values_md5sum), **kwargs)

        self._hsp_write_attribute(h5group, ('nobjects', self._nobj), **kwargs)
        self._hsp_write_attribute(h5group, ('field_value_min', self._values_min), **kwargs)
        self._hsp_write_attribute(h5group, ('field_value_max', self._values_max), **kwargs)
        self._hsp_write_attribute(h5group, ('field_value_mean', self._values_mean), **kwargs)
        self._hsp_write_attribute(h5group, ('field_value_std', self._values_std), **kwargs)

        if write_values:
            self._hsp_write_callback(str(self), **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a CatalogField object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        catalog: ``CatalogField``
            Read CatalogField instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(CatalogField, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Try to read/find protocol
        try:
            objprop_uid = uuid.UUID(cls._hsp_read_attribute(h5group, "object_property_uid", "object property UUID",
                                                            raise_error_if_not_found=True))

            # Search for already instantiated object property in dependency object dictionary
            if dependency_objdict is None or ObjectProperty.__name__ not in dependency_objdict:
                err_msg = "Cannot find any object property already instantiated in the project."
                log.error(err_msg)
                raise IOError(err_msg)

            # Find protocol according to its UUID
            objprop_dict = dependency_objdict[ObjectProperty.__name__]
            if objprop_uid not in objprop_dict:
                err_msg = "Cannot find object property with uid {uid:s}.".format(uid=str(objprop_uid))
                log.error(err_msg)
                raise IOError(err_msg)

            obj_prop = objprop_dict[objprop_uid]
        except IOError:  # Protocol UUID not found in Catalog field
            raise

        # Create catalog field object
        cf = cls(uid=uid, obj_prop=obj_prop, hdf5_init=True)

        # Read field value array MD5 sum + stats attributes
        cf._values_md5sum = cls._hsp_read_attribute(h5group, "field_values_md5sum", "catalog field value array md5sum",
                                                    raise_error_if_not_found=True)
        cf._nobj = cls._hsp_read_attribute(h5group, "nobjects", "number of objects",
                                           raise_error_if_not_found=True)
        cf._values_min = cls._hsp_read_attribute(h5group, "field_value_min", "field min. value",
                                                 raise_error_if_not_found=True)
        cf._values_max = cls._hsp_read_attribute(h5group, "field_value_max", "field max. value",
                                                 raise_error_if_not_found=True)
        cf._values_mean = cls._hsp_read_attribute(h5group, "field_value_mean", "field mean. value",
                                                  raise_error_if_not_found=True)
        cf._values_std = cls._hsp_read_attribute(h5group, "field_value_std", "field std. value",
                                                 raise_error_if_not_found=True)

        # Set HDF5 group/file info for lazy I/O
        cf._hsp_set_lazy_source(h5group)

        return cf

    def _hsp_lazy_read_data(self, h5group):
        """
        Lazy read method to load field values from HDF5 file (*.h5)

        Parameters
        ----------
        h5group: `h5py.Group`
        """
        # Read field value array
        self._val_series = self._hsp_read_dataset(h5group, "field_values", "catalog field value array",
                                                  raise_error_if_not_found=True)

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        if self._obj_property.filter_flag in [PropertyFilterFlag.BASIC_FILTER, PropertyFilterFlag.ADVANCED_FILTER] and\
                self._obj_property.datatype not in [DataType.REAL, DataType.INTEGER]:
            log.warning("Can only filter numerical fields.")

    def __unicode__(self):
        """
        String representation of the instance
        """
        return "'{pname:s}' catalog field".format(pname=self.property_name)


__all__ = ["CatalogField"]



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
from __future__ import unicode_literals
from setuptools import setup, find_packages


setup(
    name="astrophysix",
    version="0.7.0",  # Also found in astrophysix.__init__.py and in doc/source/conf.py
    install_requires=[
        'enum34;python_version<"3.4"',
        'pytz>=2019.1;python_version<"3.4"',
        'h5py>=2.10.0',
        'future>=0.17.1',
        'numpy>=1.16.4',
        'Pillow>=6.2.1',
        'pandas>=0.24.2'
    ],
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4',  # 2.7.x or 3.5/3.6/3.7/3.8/3.9/3.10
    packages=find_packages(include=['astrophysix', 'astrophysix.*']),
    package_data={'astrophysix': ["astrophysix/simdm/d3_web/plot*.html",
                                  "astrophysix/simdm/d3_web/MathJax_local.js",
                                  "astrophysix/simdm/d3_web/d3_plot.gif",
                                  "astrophysix/simdm/d3_web/loading.png",
                                  # "tests/simdm/io/*"
    ]},
    include_package_data=True,

    # Metadata to display on PyPI
    author="Damien CHAPON",
    description="Astrophysical simulation project documentation package",
    long_description="""Astrophysical simulation project documentation tool. Lets the user create scientific analysis 
                        study portable HDF5 files based on the IVOA Simulation Datamodel standard. It also allow users
                        to handle some common and most useful physical quantities and units.""",
    author_email="damien.chapon@cea.fr",
    license="CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1, http://www.cecill.info)",

    keywords="Astrophysics simulation analysis publication",
    url="https://astrophysix.readthedocs.io",
    download_url="https://gitlab.com/coast-dev/astrophysix",
    project_urls={
        "Source": "https://drf-gitlab.cea.fr/coast-dev/Astrophysix/",
    },

    classifiers=["Intended Audience :: Science/Research",
                 "License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)",
                 "Operating System :: MacOS :: MacOS X",
                 "Operating System :: POSIX :: Linux",
                 # "Development Status :: 3 - Alpha",
                 # "Development Status :: 4 - Beta",
                 "Development Status :: 5 - Production/Stable",
                 "Programming Language :: Python :: 2.7",
                 "Programming Language :: Python :: 3.6",
                 "Programming Language :: Python :: 3.7",
                 "Programming Language :: Python :: 3.8",
                 "Programming Language :: Python :: 3.9",
                 "Programming Language :: Python :: 3.10",
                 "Topic :: Software Development :: Version Control :: Git",
                 "Topic :: Scientific/Engineering :: Physics",
                 "Topic :: Scientific/Engineering :: Astronomy",
                 "Topic :: Scientific/Engineering :: Information Analysis"],
    entry_points={
        "console_scripts": [
            "gald3_server = astrophysix.simdm.d3_web.__main__:main"
        ]
    },
)



# -*- coding: utf-8 -*-
# This file is part of the 'astrophysix' Python package.
#
# Copyright © Commissariat a l'Energie Atomique et aux Energies Alternatives (CEA)
#
#  FREE SOFTWARE LICENCING
#  -----------------------
# This software is governed by the CeCILL license under French law and abiding by the rules of distribution of free
# software. You can use, modify and/or redistribute the software under the terms of the CeCILL license as circulated by
# CEA, CNRS and INRIA at the following URL: "http://www.cecill.info". As a counterpart to the access to the source code
# and rights to copy, modify and redistribute granted by the license, users are provided only with a limited warranty
# and the software's author, the holder of the economic rights, and the successive licensors have only limited
# liability. In this respect, the user's attention is drawn to the risks associated with loading, using, modifying
# and/or developing or reproducing the software by the user in light of its specific status of free software, that may
# mean that it is complicated to manipulate, and that also therefore means that it is reserved for developers and
# experienced professionals having in-depth computer knowledge. Users are therefore encouraged to load and test the
# software's suitability as regards their requirements in conditions enabling the security of their systems and/or data
# to be ensured and, more generally, to use and operate it in the same conditions as regards security. The fact that
# you are presently reading this means that you have had knowledge of the CeCILL license and that you accept its terms.
#
#
# COMMERCIAL SOFTWARE LICENCING
# -----------------------------
# You can obtain this software from CEA under other licencing terms for commercial purposes. For this you will need to
# negotiate a specific contract with a legal representative of CEA.
#
"""
@startuml


AssociatedFile <|-- HDF5File
AssociatedFile <|-- AsciiFile
AssociatedFile <|-- CSVFile
AssociatedFile <|-- JsonFile
AssociatedFile <|-- TarGzFile
AssociatedFile <|-- PickleFile
AssociatedFile <|-- YamlFile
AssociatedFile <|-- FitsFile
AssociatedFile <|-- ImageFile

ImageFile <|-- PngImageFile
ImageFile <|-- JpegImageFile


Datafile "1" *-- "many" AssociatedFile : contains

class FileType {
    {field}+PNG_FILE
    {field}+HDF5_FILE
    {field}+XML_FILE
    {field}+PICKLE_FILE
    {field}+ASCII_FILE
    {field}+JSON_FILE
    {field}+PNG_FILE
    {field}+JPEG_FILE
    {field}+FITS_FILE
    {field}+YAML_FILE
    {field}+TARGZ_FILE
    {field}+CSV_FILE

    {method} +alias
    {method} +extension_list
    {method} +from_alias()
    {method} +default_extension
}


class ImageFile {
    +pil_image: Pillow Image
}

class AssociatedFile {
    +filename : string
    +file_md5sum : string
    +last_modified : datetime
    +raw_file_data : bytes
}

@enduml
"""
from __future__ import print_function, unicode_literals, division, absolute_import  # Python 2 and 3 compatibility
from future.builtins import str, list, int
import logging
import numpy as N

from astrophysix.utils.file import FileType
from astrophysix.utils.strings import Stringifiable
from astrophysix.utils.persistency import Hdf5StudyPersistent
from .plot import PlotInfo
from .file import AssociatedFile
from ..utils import GalacticaValidityCheckMixin

__doc__ = """

.. autoclass:: astrophysix.simdm.datafiles.Datafile
    :members:
    :undoc-members:
    :inherited-members:
    :exclude-members: open_h5file, is_type_string, hsp_save_to_h5, cast_string, hsp_load_from_h5, 
                      INVALID_ALIAS_ERROR_MESSAGE, VALID_ALIAS_REGEX

"""

log = logging.getLogger("astrophysix.simdm")


class Datafile(Hdf5StudyPersistent, GalacticaValidityCheckMixin, Stringifiable):
    """
    Datafile class

    Parameters
    ----------
    name: :obj:`string`
        datafile name (mandatory)
    description: :obj:`string`
        datafile description

    Example
    -------
    >>> from astrophysix.utils import FileType
    >>> from astrophysix.simdm.datafiles import JpegImageFile
    >>> df = Datafile(name="Pre-stellar cores mass spectrum")
    >>> df[FileType.PNG_FILE] = "/data/SIMUS/result_spectrum/mass_spectrum.png"
    >>> df[FileType.FITS_FILE] = "/data/SIMUS/result_spectrum/pre-stellar-core-mass-hist.fits"
    >>> df[FileType.PNG_FILE] = JpegImageFile.load_file("/data/SIMUS/result_spectrum/hist.jpg")
    ValueError: Datafile associated file type mismatch : expected PngImageFile object but JpegImageFile was provided.
    >>> df[FileType.PNG_FILE] = "/data/SIMUS/result_spectrum/hist.jpg"
    AttributeError: Invalid filename for a PNG file (/data/SIMUS/result_spectrum/hist.jpg).
    >>> # Removing a file
    >>> del df[FileType.FITS_FILE]
    """
    def __init__(self, **kwargs):
        super(Datafile, self).__init__(**kwargs)
        self._name = ""
        self._desc = ""
        self._files = {}  # Datafile associated files dictionary, indexed by FileType.
        self._plot_info = None

        # Datafile name
        if "name" not in kwargs:
            raise AttributeError("{cname:s} 'name' attribute is not defined (mandatory).".format(cname=self.__class__.__name__))
        self.name = kwargs["name"]

        if "description" in kwargs:
            self.description = kwargs["description"]

    def __eq__(self, other):
        """
        Datafile comparison method

        Parameters
        ----------
        other: :class:`~astrophysix.simdm.datafiles.Datafile`
            datafile to compare to:
        """
        if not super(Datafile, self).__eq__(other):
            return False

        if self._name != other.name:
            return False

        if self._desc != other.description:
            return False

        # Lazy
        if self.plot_info != other.plot_info:
            return False

        # Datafile comparison
        for ft in FileType:
            try:
                af = self[ft]
            except KeyError:
                af = None
            try:
                oaf = other[ft]
            except KeyError:
                oaf = None
            if oaf is None and af is None:
                continue
            if oaf != af:
                return False

        return True

    @property
    def name(self):
        """Datafile name. Can be set to a non-empty :obj:`string` value."""
        return self._name

    @name.setter
    def name(self, new_df_name):
        try:
            self._name = Stringifiable.cast_string(new_df_name, valid_empty=False)
        except TypeError:
            err_msg = "{cname:s} 'name' property is not a valid (non-empty) string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    @property
    def description(self):
        """Datafile description. Can be set to any :obj:`string` value."""
        return self._desc

    @description.setter
    def description(self, new_descr):
        try:
            self._desc = Stringifiable.cast_string(new_descr)
        except TypeError:
            err_msg = "{cname:s} 'description' property is not a valid string.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)

    def __getitem__(self, ftype):
        """
        Get an associated file from the data file, given its file type.

        Parameters
        ----------
        ftype: :class:`~astrophysix.utils.file.FileType`
            Associated file type

        Returns
        -------
        f: :class:`~astrophysix.simdm.datafiles.file.AssociatedFile`
            datafile associated file for the required file type.

        Raises
        ------
        KeyError
            if the search index type is not a :class:`~astrophysix.utils.file.FileType` instance or if there is no
            associated file with the required file type.
        """
        if not isinstance(ftype, FileType):
            err_msg = "{cname:s} '{ft!s}' key is not a valid {ftcname:s} " \
                      "index.".format(cname=self.__class__.__name__, ft=ftype, ftcname=FileType.__name__)
            log.error(err_msg)
            raise KeyError(err_msg)

        f = self._files.get(ftype, None)
        if f is None:
            err_msg = "No {ft:s} file is associated to the {df!s}.".format(ft=ftype.alias, df=self)
            log.error(err_msg)
            raise KeyError(err_msg)

        return f

    def __delitem__(self, ftype):
        """
        Remove associated file given its file type.

        Parameters
        ----------
        item: :class:`~astrophysix.utils.file.FileType`

        Raises
        ------
        KeyError
            if the search index type is not a :class:`~astrophysix.utils.file.FileType` instance or if there is no
            associated file with the required file type.
        """
        if not isinstance(ftype, FileType):
            err_msg = "{cname:s} '{ft!s}' key is not a valid {ftcname:s} " \
                      "index.".format(cname=self.__class__.__name__, ft=ftype, ftcname=FileType.__name__)
            log.error(err_msg)
            raise KeyError(err_msg)

        if ftype not in self._files:
            err_msg = "No {ft:s} file is associated to the {df!s}.".format(ft=ftype.alias, df=self)
            log.error(err_msg)
            raise KeyError(err_msg)

        del self._files[ftype]

    def __setitem__(self, filetype, ass_file):
        """
        Set an associated file with a given file type into the data file.

        Parameters
        ----------
        filetype: :class:`~astrophysix.utils.file.FileType`
            Associated file type
        ass_file: :obj:`string` or :class:`~astrophysix.simdm.datafiles.file.AssociatedFile`
            Associated file path or instance 
        """
        if not isinstance(filetype, FileType):
            err_msg = "{cname:s} '{ft!s}' key is not a valid {ftcname:s} " \
                      "index.".format(cname=self.__class__.__name__, ft=filetype, ftcname=FileType.__name__)
            log.error(err_msg)
            raise KeyError(err_msg)

        for kl in AssociatedFile._all_subclasses():
            if kl.FILE_TYPE == filetype:
                if Stringifiable.is_type_string(ass_file):
                    self._files[filetype] = kl.load_file(ass_file)
                elif isinstance(ass_file, AssociatedFile):
                    if not isinstance(ass_file, kl):
                        err_msg = "{cname:s} associated file type mismatch : expected {kname:s} object but " \
                                  "{afc:s} was provided.".format(cname=self.__class__.__name__, kname=kl.__name__,
                                                                 afc=type(ass_file).__name__)
                        log.error(err_msg)
                        raise ValueError(err_msg)
                    self._files[filetype] = ass_file
                else:
                    err_msg = "Only file paths or {kname:s} objects can be set in {cname:s} as {ft:s} " \
                              "files.".format(cname=self.__class__.__name__, kname=kl.__name__, ft=filetype.alias)
                    log.error(err_msg)
                    raise ValueError(err_msg)
                return

        err_msg = "Cannot attach {ft:s} file to datafile !".format(ft=filetype.alias)
        log.error(err_msg)
        raise NotImplementedError(err_msg)

    @property
    def plot_info(self):
        """Datafile plot information. Can be set to a :class:`~astrophysix.simdm.datafiles.plot.PlotInfo` instance."""
        return self._plot_info

    @plot_info.setter
    def plot_info(self, new_pi):
        if not isinstance(new_pi, PlotInfo):
            err_msg = "{cname:s} 'plot_info' property is not a valid PlotInfo " \
                      "object.".format(cname=self.__class__.__name__)
            log.error(err_msg)
            raise AttributeError(err_msg)
        self._plot_info = new_pi

    def __iter__(self):
        for ft, f in self._files.items():
            yield ft, f

    def display_files(self):
        """
        Show tabulated view of associated files

        Example
        -------
            >>> df.display_files()
            [My best datafile] datafile. Attached files :
            +-----------+-----------------------------+
            | File type |          Filename           |
            +-----------+-----------------------------+
            | PNG       | CEA.png                     |
            +-----------+-----------------------------+
            | JPEG      | irfu_simple.jpg             |
            +-----------+-----------------------------+
            | FITS      | cassiopea_A_0.5-1.5keV.fits |
            +-----------+-----------------------------+
            | TARGZ     | archive.tar.gz              |
            +-----------+-----------------------------+
            | JSON      | test_header_249.json        |
            +-----------+-----------------------------+
            | YAML      | config.yml                  |
            +-----------+-----------------------------+
            | ASCII     | abstract.txt                |
            +-----------+-----------------------------+
            | HDF5      | study.h5                    |
            +-----------+-----------------------------+
            | PICKLE    | dict_saved.pkl              |
            +-----------+-----------------------------+
        """
        # Empty file dict
        if len(self._files) == 0:
            print(str(self) + " (no attachment)")
            return

        # Display associated file list in a pretty-formatted table
        strrep = str(self) + ". Attached files :\n"
        ft_fname_tuplist = list([(ft.alias, f.filename) for ft, f in self._files.items()])
        nft = N.max([len(t[0]) for t in ft_fname_tuplist])
        ind_header = "File type"
        nft = len(ind_header) if nft < len(ind_header) else nft
        nstr = N.max([len(t[1]) for t in ft_fname_tuplist])
        item_header = "Filename"
        nstr = len(item_header) if nstr < len(item_header) else nstr
        interline = "+-{nind:s}-+-{nstr:s}-+".format(nind=nft*"-", nstr=nstr*"-")
        strrep += interline + "\n| {ind:^{nind}s} | {s:^{nstr}s} |\n".format(nind=nft, ind=ind_header, nstr=nstr,
                                                                             s=item_header)
        for t in ft_fname_tuplist:
            strrep += interline + "\n"
            strrep += "| {ind:<{nind}s} | {s:<{nstr}s} |\n".format(nind=nft, ind=t[0], nstr=nstr, s=t[1])
        strrep += interline

        print(strrep)

    @classmethod
    def _hsp_valid_attributes(cls):
        """List of valid kwargs in __init__() method"""
        return ["name", "description"]

    def _hsp_write(self, h5group, **kwargs):
        """
        Serialize a Datafile object into a HDF5 file.

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to write the Datafile into.
        kwargs: ``dict``
            keyword argument dictionary.
        """
        # Call to parent class _hsp_write() : write UUID, etc.
        super(Datafile, self)._hsp_write(h5group, **kwargs)

        # Write name
        self._hsp_write_attribute(h5group, ('name', self._name), **kwargs)

        # Write description, if defined
        self._hsp_write_attribute(h5group, ('description', self._desc), **kwargs)

        # Write plot info, if defined
        self._hsp_write_object(h5group, "PLOT_INFO", self._plot_info, **kwargs)

        # Write associated files into HDF5 file
        if not kwargs.get("dry_run", False) and not kwargs.get("new_file", True):  # Old HDF5 file being modified (not a dry run)
            # Delete files from HDF5 group if not present anymore in the datafile
            for ft in FileType:
                if ft.alias in h5group and ft not in self._files:
                    del h5group[ft.alias]

        for ft, f in self._files.items():
            self._hsp_write_object(h5group, ft.alias, f, **kwargs)

    @classmethod
    def _hsp_read(cls, h5group, version, dependency_objdict=None):
        """
        Read a Datafile object from a HDF5 file (*.h5).

        Parameters
        ----------
        h5group: ``h5py.Group``
            Main group to read the object from.
        version: ``int``
            version of the object to read.
        dependency_objdict: ``dict``
            dependency object dictionary. Default None

        Returns
        -------
        df: ``Datafile``
            Read Datafile instance
        """
        # Handle different versions here

        # Fetch Hdf5StudyPersistent object UUID
        uid = super(Datafile, cls)._hsp_read(h5group, version, dependency_objdict=dependency_objdict)

        # Read datafile name
        name = cls._hsp_read_attribute(h5group, 'name', "datafile name")

        # Create datafile object
        df = cls(uid=uid, name=name)

        # Read datafile description, if defined
        df_descr = cls._hsp_read_attribute(h5group, 'description', "datafile description",
                                           raise_error_if_not_found=False)
        if df_descr is not None:
            df.description = df_descr

        # Read plot info
        if "PLOT_INFO" in h5group:
            df.plot_info = PlotInfo._hsp_read_object(h5group, "PLOT_INFO", "plot information")

        # Read associated files from HDF5
        for ft in FileType:
            if ft.alias not in h5group:
                continue
            df[ft] = AssociatedFile._hsp_read_object(h5group, ft.alias, "{ft:s} associated file".format(ft=ft.alias))

        return df

    def galactica_validity_check(self, **kwargs):
        """
        Perform validity checks on this  instance and eventually log warning messages.

        Parameters
        ----------
        kwargs: :obj:`dict`
            keyword arguments (optional)
        """
        # Check datafile name
        if len(self._name) > 128:
            log.warning("{p!s} name is too long for Galactica (max. 128 characters).".format(p=self))

        # Check associated files validity
        for ft, af in self._files.items():
            af.galactica_validity_check(**kwargs)

        # Check plot info validity, if defined
        if self._plot_info is not None:
            self._plot_info.galactica_validity_check(**kwargs)

    def __unicode__(self):
        """
        String representation of the data file instance
        """
        s = "[{df_name:s}] datafile".format(df_name=self._name)
        return s


__all__ = ["Datafile"]
